# solr_collection.py

import logging
from copy import copy
from typing import Any, Dict, List, Optional, Union

import requests

from linkml_store.api import Collection
from linkml_store.api.collection import DEFAULT_FACET_LIMIT
from linkml_store.api.queries import Query, QueryResult

logger = logging.getLogger(__name__)


class SolrCollection(Collection):

    @property
    def _collection_base(self) -> str:
        if self.parent.use_cores:
            base_url = f"{self.parent.base_url}/{self.alias}"
        else:
            base_url = self.parent.base_url
        return base_url

    def search(
        self,
        query: str,
        where: Optional[Any] = None,
        index_name: Optional[str] = None,
        limit: Optional[int] = None,
        **kwargs,
    ) -> QueryResult:
        if index_name is None:
            index_name = "edismax"
        qfs = self.parent.metadata.searchable_slots
        if not qfs:
            raise ValueError("No searchable slots configured for Solr collection")
        solr_query = self._build_solr_query(where, search_term=query, extra={"defType": index_name, "qf": qfs})
        logger.info(f"Querying Solr collection {self.alias} with query: {solr_query}")

        response = requests.get(f"{self._collection_base}/select", params=solr_query)
        response.raise_for_status()

        data = response.json()
        num_rows = data["response"]["numFound"]
        rows = data["response"]["docs"]
        ranked_rows = [(1.0, row) for row in rows]
        return QueryResult(query=where, search_term=query, num_rows=num_rows, rows=rows, ranked_rows=ranked_rows)

    def query(self, query: Query, **kwargs) -> QueryResult:
        solr_query = self._build_solr_query(query)
        logger.info(f"Querying Solr collection {self.alias} with query: {solr_query}")

        response = requests.get(f"{self._collection_base}/select", params=solr_query)
        response.raise_for_status()

        data = response.json()
        num_rows = data["response"]["numFound"]
        rows = data["response"]["docs"]

        return QueryResult(query=query, num_rows=num_rows, rows=rows)

    def query_facets(
        self, where: Optional[Dict] = None, facet_columns: List[str] = None, facet_limit=DEFAULT_FACET_LIMIT, **kwargs
    ) -> Dict[str, Dict[str, int]]:
        solr_query = self._build_solr_query(where)
        solr_query["facet"] = "true"
        solr_query["facet.field"] = facet_columns
        solr_query["facet.limit"] = facet_limit

        logger.info(f"Querying Solr collection {self.alias} for facets with query: {solr_query}")

        response = requests.get(f"{self._collection_base}/select", params=solr_query)
        response.raise_for_status()

        data = response.json()
        facet_counts = data["facet_counts"]["facet_fields"]

        results = {}
        for facet_field, counts in facet_counts.items():
            results[facet_field] = list(zip(counts[::2], counts[1::2]))

        return results

    def _build_solr_query(
        self, query: Union[Query, Dict], search_term="*:*", extra: Optional[Dict] = None
    ) -> Dict[str, Any]:
        solr_query = {}
        if query is None:
            query = {}

        if isinstance(query, Query):
            where = query.where_clause
            solr_query["fq"] = self._build_solr_where_clause(where)

            if query.select_cols:
                solr_query["fl"] = ",".join(query.select_cols)

            if query.limit:
                solr_query["rows"] = query.limit

            if query.offset:
                solr_query["start"] = query.offset

        elif isinstance(query, dict):
            solr_query["fq"] = self._build_solr_where_clause(query)

        solr_query["wt"] = "json"
        if "q" not in solr_query:
            solr_query["q"] = search_term
        if extra:
            solr_query.update(extra)
        logger.info(f"Built Solr query: {solr_query}")
        return solr_query

    def _build_solr_where_clause(self, where_clause: Dict) -> str:
        if where_clause is None:
            where_clause = {}
        conditions = []
        if self.parent.metadata.collection_type_slot:
            where_clause = copy(where_clause)
            where_clause[self.parent.metadata.collection_type_slot] = self.alias
        for field, value in where_clause.items():
            if not isinstance(value, (list, tuple)):
                value = [value]
            value = [f'"{v}"' if isinstance(v, str) else str(v) for v in value]
            if len(value) > 1:
                conditions.append(f"{field}:({' '.join(value)})")
            else:
                conditions.append(f"{field}:{value[0]}")

        return " AND ".join(conditions)



# neo4j_database.py

import logging
from pathlib import Path
from typing import Optional, Union

from neo4j import Driver, GraphDatabase, Session

from linkml_store.api import Database
from linkml_store.api.queries import Query, QueryResult
from linkml_store.api.stores.neo4j.neo4j_collection import Neo4jCollection
from linkml_store.utils.format_utils import Format

logger = logging.getLogger(__name__)


class Neo4jDatabase(Database):
    """
    An adapter for Neo4j databases.
    """

    _driver: Driver = None
    collection_class = Neo4jCollection

    def __init__(self, handle: Optional[str] = None, **kwargs):
        # Note: in the community editing the database must be "neo4j"
        if handle is None:
            handle = "bolt://localhost:7687/neo4j"
        if handle.startswith("neo4j:"):
            handle = handle.replace("neo4j:", "bolt:")
        super().__init__(handle=handle, **kwargs)

    @property
    def _db_name(self) -> str:
        if self.handle:
            db = self.handle.split("/")[-1]
        else:
            db = "default"
        return db

    @property
    def driver(self) -> Driver:
        if self._driver is None:
            uri, user, password = self._parse_handle()
            self._driver = GraphDatabase.driver(uri, auth=(user, password))
        return self._driver

    def session(self) -> Session:
        return self.driver.session(database=self._db_name)

    def _parse_handle(self):
        parts = self.handle.split("://")
        protocol = parts[0]
        rest = parts[1]

        if "@" in rest:
            auth, host = rest.split("@")
            user, password = auth.split(":")
        else:
            host = rest
            user, password = "neo4j", "password"  # Default credentials

        uri = f"{protocol}://{host}"
        return uri, user, password

    def commit(self, **kwargs):
        # Neo4j uses auto-commit by default for each transaction
        pass

    def close(self, **kwargs):
        if self._driver:
            self._driver.close()

    def drop(self, **kwargs):
        with self.driver.session() as session:
            session.run("MATCH (n) DETACH DELETE n")

    def query(self, query: Query, **kwargs) -> QueryResult:
        if query.from_table:
            collection = self.get_collection(query.from_table)
            return collection.query(query, **kwargs)
        else:
            raise NotImplementedError(f"Querying without a table is not supported in {self.__class__.__name__}")

    def init_collections(self):
        if self._collections is None:
            self._collections = {}

        # In Neo4j, we don't have a direct equivalent to collections
        # We'll use node labels as a proxy for collections
        with self.driver.session() as session:
            result = session.run("CALL db.labels()")
            labels = [record["label"] for record in result]

            for label in labels:
                if label not in self._collections:
                    collection = Neo4jCollection(name=label, parent=self)
                    self._collections[label] = collection

    def export_database(self, location: str, target_format: Optional[Union[str, Format]] = None, **kwargs):
        # Neo4j doesn't have a built-in export function, so we'll implement a basic JSON export
        if target_format == Format.JSON or target_format == "json":
            path = Path(location)
            with self.driver.session() as session:
                result = session.run("MATCH (n) RETURN n")
                nodes = [dict(record["n"].items()) for record in result]

                result = session.run("MATCH ()-[r]->() RETURN r")
                relationships = [
                    {
                        "type": record["r"].type,
                        "start": record["r"].start_node.id,
                        "end": record["r"].end_node.id,
                        **dict(record["r"].items()),
                    }
                    for record in result
                ]

                data = {"nodes": nodes, "relationships": relationships}

                import json

                with open(path, "w") as f:
                    json.dump(data, f)
        else:
            super().export_database(location, target_format=target_format, **kwargs)

    def import_database(self, location: str, source_format: Optional[str] = None, **kwargs):
        if source_format == Format.JSON or source_format == "json":
            path = Path(location)
            with open(path, "r") as f:
                import json

                data = json.load(f)

            with self.driver.session() as session:
                for node in data["nodes"]:
                    labels = node.pop("labels", ["Node"])
                    props = ", ".join([f"{k}: ${k}" for k in node.keys()])
                    query = f"CREATE (n:{':'.join(labels)} {{{props}}})"
                    session.run(query, **node)

                for rel in data["relationships"]:
                    # rel_type = rel.pop("type")
                    start = rel.pop("start")
                    end = rel.pop("end")
                    # props = ", ".join([f"{k}: ${k}" for k in rel.keys()])
                    query = (
                        f"MATCH (a), (b) WHERE id(a) = {start} AND id(b) = {end} "
                        "CREATE (a)-[r:{rel_type} {{{props}}}]->(b)"
                    )
                    session.run(query, **rel)
        else:
            super().import_database(location, source_format=source_format, **kwargs)



import csv
import gzip
import io
import json
import logging
import sys
import tarfile
from enum import Enum
from io import StringIO
from pathlib import Path
from typing import IO, Any, Dict, List, Optional, TextIO, Type, Union

import pandas as pd
import pystow
import yaml
from pydantic import BaseModel
from tabulate import tabulate

logger = logging.getLogger(__name__)


class Format(Enum):
    """
    Supported generic file formats for loading and rendering objects.
    """

    JSON = "json"
    JSONL = "jsonl"
    YAML = "yaml"
    TSV = "tsv"
    CSV = "csv"
    PYTHON = "python"
    PARQUET = "parquet"
    FORMATTED = "formatted"
    TABLE = "table"
    SQLDUMP_DUCKDB = "duckdb"
    SQLDUMP_POSTGRES = "postgres"
    DUMP_MONGODB = "mongodb"

    @classmethod
    def guess_format(cls, file_name: str) -> Optional["Format"]:
        ext = Path(file_name).suffix.lower()

        format_map = {
            ".json": cls.JSON,
            ".jsonl": cls.JSONL,
            ".yaml": cls.YAML,
            ".yml": cls.YAML,
            ".tsv": cls.TSV,
            ".csv": cls.CSV,
            ".py": cls.PYTHON,
            ".parquet": cls.PARQUET,
            ".pq": cls.PARQUET,
        }
        fmt = format_map.get(ext, None)
        if fmt is None:
            if ext.startswith("."):
                ext = ext[1:]
            if ext in [f.value for f in Format]:
                return Format(ext)
        return fmt

    def is_dump_format(self):
        return self in [Format.SQLDUMP_DUCKDB, Format.SQLDUMP_POSTGRES, Format.DUMP_MONGODB]


def load_objects_from_url(
    url: str,
    format: Union[Format, str] = None,
    expected_type: Type = None,
    local_path: Optional[str] = None,
    **kwargs,
) -> List[Dict[str, Any]]:
    """
    Load objects from a URL in JSON, JSONLines, YAML, CSV, or TSV format.

    :param url: The URL to the file.
    :param format: The format of the file. Can be a Format enum or a string value.
    :param expected_type: The target type to load the objects into.
    :param local_path: The local path to save the file to.
    :return: A list of dictionaries representing the loaded objects.
    """
    local_path = pystow.ensure("linkml", "linkml-store", url=url)
    logger.info(f"synced to {local_path}")
    objs = load_objects(local_path, format=format, expected_type=expected_type, **kwargs)
    if not objs:
        raise ValueError(f"No objects loaded from URL: {url}")
    return objs


def process_file(
    f: IO, format: Format, expected_type: Optional[Type] = None, header_comment_token: Optional[str] = None
) -> List[Dict[str, Any]]:
    """
    Process a single file and return a list of objects.
    """
    if format == Format.JSON:
        objs = json.load(f)
    elif format == Format.JSONL:
        objs = [json.loads(line) for line in f]
    elif format == Format.YAML:
        if expected_type and expected_type == list:  # noqa E721
            objs = list(yaml.safe_load_all(f))
        else:
            objs = yaml.safe_load(f)
    elif format in [Format.TSV, Format.CSV]:
        if header_comment_token:
            while True:
                pos = f.tell()
                line = f.readline()
                if not line.startswith(header_comment_token):
                    f.seek(pos)
                    break
        delimiter = "\t" if format == Format.TSV else ","
        reader = csv.DictReader(f, delimiter=delimiter)
        objs = list(reader)
    elif format == Format.PARQUET:
        import pyarrow.parquet as pq

        table = pq.read_table(f)
        objs = table.to_pandas().to_dict(orient="records")
    elif format in [Format.PYTHON, Format.FORMATTED, Format.TABLE]:
        raise ValueError(f"Format {format} is not supported for loading objects")
    else:
        raise ValueError(f"Unsupported file format: {format}")

    if not isinstance(objs, list):
        objs = [objs]
    return objs


def load_objects(
    file_path: Union[str, Path],
    format: Optional[Union[Format, str]] = None,
    compression: Optional[str] = None,
    expected_type: Optional[Type] = None,
    header_comment_token: Optional[str] = None,
) -> List[Dict[str, Any]]:
    """
    Load objects from a file or archive in supported formats.
    For tgz archives, it processes all files and concatenates the results.

    :param file_path: The path to the file or archive.
    :param format: The format of the file. Can be a Format enum or a string value.
    :param compression: The compression type. Supports 'gz' for gzip and 'tgz' for tar.gz.
    :param expected_type: The target type to load the objects into, e.g. list
    :param header_comment_token: Token used for header comments to be skipped
    :return: A list of dictionaries representing the loaded objects.
    """
    if isinstance(file_path, Path):
        file_path = str(file_path)

    if isinstance(format, str):
        format = Format(format)

    all_objects = []

    if compression == "tgz":
        with tarfile.open(file_path, "r:gz") as tar:
            for member in tar.getmembers():
                if member.isfile():
                    f = tar.extractfile(member)
                    if f:
                        content = io.TextIOWrapper(f)
                        member_format = Format.guess_format(member.name) if not format else format
                        logger.debug(f"Processing tar member {member.name} with format {member_format}")
                        all_objects.extend(process_file(content, member_format, expected_type, header_comment_token))
    else:
        if Path(file_path).is_dir():
            raise ValueError(f"{file_path} is a dir, which is invalid for {format}")
        mode = "rb" if format == Format.PARQUET or compression == "gz" else "r"
        open_func = gzip.open if compression == "gz" else open
        format = Format.guess_format(file_path) if not format else format
        with open_func(file_path, mode) if file_path != "-" else sys.stdin as f:
            if compression == "gz" and mode == "r":
                f = io.TextIOWrapper(f)
            all_objects = process_file(f, format, expected_type, header_comment_token)

    logger.debug(f"Loaded {len(all_objects)} objects from {file_path}")
    return all_objects


def xxxload_objects(
    file_path: Union[str, Path],
    format: Union[Format, str] = None,
    compression: Optional[str] = None,
    expected_type: Type = None,
    header_comment_token: Optional[str] = None,
) -> List[Dict[str, Any]]:
    """
    Load objects from a file in JSON, JSONLines, YAML, CSV, or TSV format.

    >>> load_objects("tests/input/test_data/data.csv")
    [{'id': '1', 'name': 'John', 'age': '30'},
     {'id': '2', 'name': 'Alice', 'age': '25'}, {'id': '3', 'name': 'Bob', 'age': '35'}]

    :param file_path: The path to the file.
    :param format: The format of the file. Can be a Format enum or a string value.
    :param expected_type: The target type to load the objects into, e.g. list
    :return: A list of dictionaries representing the loaded objects.
    """
    if isinstance(format, str):
        format = Format(format)

    if isinstance(file_path, Path):
        file_path = str(file_path)

    if not format and (file_path.endswith(".parquet") or file_path.endswith(".pq")):
        format = Format.PARQUET
    if not format and file_path.endswith(".tsv"):
        format = Format.TSV
    if not format and file_path.endswith(".csv"):
        format = Format.CSV
    if not format and file_path.endswith(".py"):
        format = Format.PYTHON

    mode = "r"
    if format == Format.PARQUET:
        mode = "rb"

    if file_path == "-":
        # set file_path to be a stream from stdin
        f = sys.stdin
    else:
        f = open(file_path, mode)

    if format == Format.JSON or (not format and file_path.endswith(".json")):
        objs = json.load(f)
    elif format == Format.JSONL or (not format and file_path.endswith(".jsonl")):
        objs = [json.loads(line) for line in f]
    elif format == Format.YAML or (not format and (file_path.endswith(".yaml") or file_path.endswith(".yml"))):
        if expected_type and expected_type == list:  # noqa E721
            objs = list(yaml.safe_load_all(f))
        else:
            objs = yaml.safe_load(f)
    elif format == Format.TSV or format == Format.CSV:
        # Skip initial comment lines if comment_char is set
        if header_comment_token:
            # Store the original position
            original_pos = f.tell()

            # Read and store lines until we find a non-comment line
            lines = []
            for line in f:
                if not line.startswith(header_comment_token):
                    break
                lines.append(line)

            # Go back to the original position
            f.seek(original_pos)

            # Skip the comment lines we found
            for _ in lines:
                f.readline()
        if format == Format.TSV:
            reader = csv.DictReader(f, delimiter="\t")
        else:
            reader = csv.DictReader(f)
        objs = list(reader)
    elif format == Format.PARQUET:
        import pyarrow.parquet as pq

        table = pq.read_table(f)
        objs = table.to_pandas().to_dict(orient="records")
    else:
        raise ValueError(f"Unsupported file format: {file_path}")
    if not isinstance(objs, list):
        objs = [objs]
    return objs


def write_output(
    data: Union[List[Dict[str, Any]], Dict[str, Any], pd.DataFrame],
    format: Union[Format, str] = Format.YAML,
    target: Optional[Union[TextIO, str, Path]] = None,
) -> None:
    """
    Write output data to a file in JSON, JSONLines, YAML, CSV, or TSV format.

    >>> write_output([{"a": 1, "b": 2}, {"a": 3, "b": 4}], Format.JSON, sys.stdout)
    [
      {
        "a": 1,
        "b": 2
      },
      {
        "a": 3,
        "b": 4
        }
    ]
    """
    output_str = render_output(data, format)
    if target:
        if isinstance(target, str):
            with open(target, "w") as target:
                target.write(output_str)
        else:
            target.write(output_str)
    else:
        print(output_str)


def render_output(
    data: Union[List[Dict[str, Any]], Dict[str, Any], pd.DataFrame], format: Optional[Union[Format, str]] = Format.YAML
) -> str:
    """
    Render output data in JSON, JSONLines, YAML, CSV, or TSV format.

    >>> print(render_output([{"a": 1, "b": 2}, {"a": 3, "b": 4}], Format.JSON))
    [
      {
        "a": 1,
        "b": 2
      },
      {
        "a": 3,
        "b": 4
      }
    ]


    :param data: The data to be rendered.
    :param format: The desired output format. Can be a Format enum or a string value.
    :return: The rendered output as a string.
    """
    if isinstance(format, str):
        format = Format(format)

    if format == Format.FORMATTED:
        if not isinstance(data, pd.DataFrame):
            data = pd.DataFrame(data)
        return str(data)

    if isinstance(data, pd.DataFrame):
        data = data.to_dict(orient="records")

    if isinstance(data, dict) and format in [Format.TSV, Format.CSV]:
        data = [data]

    if isinstance(data, BaseModel):
        data = data.model_dump()

    if format == Format.JSON:
        return json.dumps(data, indent=2, default=str)
    elif format == Format.JSONL:
        return "\n".join(json.dumps(obj) for obj in data)
    elif format == Format.PYTHON:
        return str(data)
    elif format == Format.TABLE:
        return tabulate(pd.DataFrame(data), headers="keys", tablefmt="psql")
    elif format == Format.YAML:
        if isinstance(data, list):
            return yaml.safe_dump_all(data, sort_keys=False)
        else:
            return yaml.safe_dump(data, sort_keys=False)
    elif format == Format.TSV:
        output = StringIO()
        writer = csv.DictWriter(output, fieldnames=get_fieldnames(data), delimiter="\t")
        writer.writeheader()
        writer.writerows(data)
        return output.getvalue()
    elif format == Format.CSV:
        output = StringIO()
        writer = csv.DictWriter(output, fieldnames=get_fieldnames(data))
        writer.writeheader()
        writer.writerows(data)
        return output.getvalue()
    else:
        raise ValueError(f"Unsupported output format: {format}")


def get_fieldnames(data: List[Dict[str, Any]]) -> List[str]:
    """
    Get the fieldnames of a list of dictionaries.

    >>> get_fieldnames([{"a": 1, "b": 2}, {"a": 3, "b": 4}])
    ['a', 'b']

    :param data: The list of dictionaries.
    :return: The fieldnames.
    """
    fieldnames = []
    for obj in data:
        fieldnames.extend([k for k in obj.keys() if k not in fieldnames])
    return fieldnames


def guess_format(path: str) -> Optional[Format]:
    """
    Guess the format of a file based on its extension.

    >>> guess_format("data.json")
    <Format.JSON: 'json'>
    >>> guess_format("data.jsonl")
    <Format.JSONL: 'jsonl'>
    >>> guess_format("data.yaml")
    <Format.YAML: 'yaml'>
    >>> assert not guess_format("data")

    :param path: The path to the file.
    :return: The guessed format.
    """
    return Format.guess_format(path)



import logging
from enum import Enum
from typing import Any, Callable, Dict, List, Optional, Tuple

import numpy as np
from pydantic import BaseModel

INDEX_ITEM = np.ndarray

logger = logging.getLogger(__name__)


class TemplateSyntaxEnum(str, Enum):
    """
    Template syntax types.
    """

    jinja2 = "jinja2"
    fstring = "fstring"


def cosine_similarity(vector1, vector2) -> float:
    """
    Calculate the cosine similarity between two vectors

    :param vector1:
    :param vector2:
    :return:
    """
    dot_product = np.dot(vector1, vector2)
    norm1 = np.linalg.norm(vector1)
    norm2 = np.linalg.norm(vector2)
    return dot_product / (norm1 * norm2)


class Indexer(BaseModel):
    """
    An indexer operates on a collection in order to search for objects.
    """

    name: Optional[str] = None
    index_type: Optional[str] = None
    index_function: Optional[Callable] = None
    distance_function: Optional[Callable] = None
    index_attributes: Optional[List[str]] = None
    text_template: Optional[str] = None
    text_template_syntax: Optional[TemplateSyntaxEnum] = None
    filter_nulls: Optional[bool] = True
    vector_default_length: Optional[int] = 1000
    index_field: Optional[str] = "__index__"

    def object_to_vector(self, obj: Dict[str, Any]) -> INDEX_ITEM:
        """
        Convert an object to an indexable object

        :param obj:
        :return:
        """
        return self.text_to_vector(self.object_to_text(obj))

    def objects_to_vectors(self, objs: List[Dict[str, Any]]) -> List[INDEX_ITEM]:
        """
        Convert a list of objects to indexable objects

        :param objs:
        :return: list of vectors
        """
        return self.texts_to_vectors([self.object_to_text(obj) for obj in objs])

    def texts_to_vectors(self, texts: List[str], cache: bool = None, **kwargs) -> List[INDEX_ITEM]:
        """
        Convert a list of texts to indexable objects

        :param texts:
        :return:
        """
        return [self.text_to_vector(text, cache=cache, **kwargs) for text in texts]

    def text_to_vector(self, text: str, cache: bool = None, **kwargs) -> INDEX_ITEM:
        """
        Convert a text to an indexable object

        :param text:
        :param cache:
        :return:
        """
        raise NotImplementedError

    def object_to_text(self, obj: Dict[str, Any]) -> str:
        """
        Convert an object to a text representation

        :param obj:
        :return:
        """
        if self.index_attributes:
            if len(self.index_attributes) == 1 and not self.text_template:
                return str(obj[self.index_attributes[0]])
            obj = {k: v for k, v in obj.items() if k in self.index_attributes}
        if self.filter_nulls:
            obj = {k: v for k, v in obj.items() if v is not None}
        if self.text_template:
            syntax = self.text_template_syntax
            if not syntax:
                if "{%" in self.text_template or "{{" in self.text_template:
                    logger.info("Detected Jinja2 syntax in text template")
                    syntax = TemplateSyntaxEnum.jinja2
            if not syntax:
                syntax = TemplateSyntaxEnum.fstring
            if syntax == TemplateSyntaxEnum.jinja2:
                from jinja2 import Template

                template = Template(self.text_template)
                return template.render(**obj)
            elif syntax == TemplateSyntaxEnum.fstring:
                return self.text_template.format(**obj)
            else:
                raise NotImplementedError(f"Cannot handle template syntax: {syntax}")
        return str(obj)

    def search(
        self, query: str, vectors: List[Tuple[str, INDEX_ITEM]], limit: Optional[int] = None
    ) -> List[Tuple[float, Any]]:
        """
        Search the index for a query string

        :param query: The query string to search for
        :param vectors: A list of indexed items, where each item is a tuple of (id, vector)
        :param limit: The maximum number of results to return (optional)
        :return: A list of item IDs or objects that match the query
        """

        # Convert the query string to a vector
        query_vector = self.text_to_vector(query, cache=False)

        distances = []

        # Iterate over each indexed item
        for item_id, item_vector in vectors:
            # Calculate the Euclidean distance between the query vector and the item vector
            # distance = 1-np.linalg.norm(query_vector - item_vector)
            distance = cosine_similarity(query_vector, item_vector)
            distances.append((distance, item_id))

        # Sort the distances in ascending order
        distances.sort(key=lambda x: -x[0])

        # Limit the number of results if specified
        if limit is not None:
            distances = distances[:limit]

        return distances



import networkx as nx
from py2neo import Graph


def draw_neo4j_graph(handle="bolt://localhost:7687", auth=("neo4j", None)):
    # Connect to Neo4j
    graph = Graph(handle, auth=auth)

    # Run a Cypher query
    query = """
    MATCH (n)-[r]->(m)
    RETURN n, r, m
    LIMIT 100
    """
    result = graph.run(query)

    # Create a NetworkX graph
    G = nx.DiGraph()  # Use DiGraph for directed edges
    for record in result:
        n = record["n"]
        m = record["m"]
        r = record["r"]
        G.add_node(n["name"], label=list(n.labels or ["-"])[0])
        G.add_node(m["name"], label=list(m.labels or ["-"])[0])
        G.add_edge(n["name"], m["name"], type=type(r).__name__)

    # Draw the graph
    pos = nx.spring_layout(G)

    # Draw nodes
    nx.draw_networkx_nodes(G, pos, node_color="lightblue", node_size=10000)

    # Draw edges
    nx.draw_networkx_edges(G, pos, edge_color="gray", arrows=True)

    # Add node labels
    node_labels = nx.get_node_attributes(G, "label")
    nx.draw_networkx_labels(G, pos, {node: f"{node}\n({label})" for node, label in node_labels.items()}, font_size=16)

    # Add edge labels
    edge_labels = nx.get_edge_attributes(G, "type")
    nx.draw_networkx_edge_labels(G, pos, edge_labels, font_size=16)



import logging
from typing import Any, Dict, List, Optional, Union

import sqlalchemy as sqla
from linkml_runtime.linkml_model import ClassDefinition, SlotDefinition
from sqlalchemy import Column, Table, delete, insert, inspect, text
from sqlalchemy.sql.ddl import CreateTable

from linkml_store.api import Collection
from linkml_store.api.collection import DEFAULT_FACET_LIMIT, OBJECT
from linkml_store.api.queries import Query
from linkml_store.api.stores.duckdb.mappings import TMAP
from linkml_store.utils.sql_utils import facet_count_sql

logger = logging.getLogger(__name__)


class DuckDBCollection(Collection):
    _table_created: bool = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def insert(self, objs: Union[OBJECT, List[OBJECT]], **kwargs):
        logger.debug(f"Inserting {len(objs)}")
        if not isinstance(objs, list):
            objs = [objs]
        if not objs:
            return
        cd = self.class_definition()
        if not cd:
            logger.debug(f"No class definition defined for {self.alias} {self.target_class_name}; will induce")
            cd = self.induce_class_definition_from_objects(objs)
        self._create_table(cd)
        table = self._sqla_table(cd)
        logger.info(f"Inserting into: {self.alias} // T={table.name}")
        engine = self.parent.engine
        col_names = [c.name for c in table.columns]
        objs = [{k: obj.get(k, None) for k in col_names} for obj in objs]
        with engine.connect() as conn:
            with conn.begin():
                conn.execute(insert(table), objs)
            conn.commit()
        self._post_insert_hook(objs)

    def delete(self, objs: Union[OBJECT, List[OBJECT]], **kwargs) -> Optional[int]:
        if not isinstance(objs, list):
            objs = [objs]
        cd = self.class_definition()
        if not cd:
            cd = self.induce_class_definition_from_objects(objs)
        table = self._sqla_table(cd)
        engine = self.parent.engine
        with engine.connect() as conn:
            for obj in objs:
                conditions = [table.c[k] == v for k, v in obj.items() if k in cd.attributes]
                stmt = delete(table).where(*conditions)
                stmt = stmt.compile(engine)
                conn.execute(stmt)
                conn.commit()
        return

    def delete_where(self, where: Optional[Dict[str, Any]] = None, missing_ok=True, **kwargs) -> Optional[int]:
        logger.info(f"Deleting from {self.target_class_name} where: {where}")
        if where is None:
            where = {}
        cd = self.class_definition()
        if not cd:
            logger.info(f"No class definition found for {self.target_class_name}, assuming not prepopulated")
            return 0
        table = self._sqla_table(cd)
        engine = self.parent.engine
        inspector = inspect(engine)
        table_exists = table.name in inspector.get_table_names()
        if not table_exists:
            logger.info(f"Table {table.name} does not exist, assuming no data")
            return 0
        with engine.connect() as conn:
            conditions = [table.c[k] == v for k, v in where.items()]
            stmt = delete(table).where(*conditions)
            stmt = stmt.compile(engine)
            result = conn.execute(stmt)
            deleted_rows_count = result.rowcount
            if deleted_rows_count == 0 and not missing_ok:
                raise ValueError(f"No rows found for {where}")
            conn.commit()
            return deleted_rows_count if deleted_rows_count > -1 else None

    def query_facets(
        self, where: Dict = None, facet_columns: List[str] = None, facet_limit=DEFAULT_FACET_LIMIT, **kwargs
    ) -> Dict[str, Dict[str, int]]:
        results = {}
        cd = self.class_definition()
        with self.parent.engine.connect() as conn:
            if not facet_columns:
                if not cd:
                    raise ValueError(f"No class definition found for {self.target_class_name}")
                facet_columns = list(cd.attributes.keys())
            for col in facet_columns:
                logger.debug(f"Faceting on {col}")
                if isinstance(col, tuple):
                    sd = SlotDefinition(name="PLACEHOLDER")
                else:
                    sd = cd.attributes[col]
                facet_query = self._create_query(where_clause=where)
                facet_query_str = facet_count_sql(facet_query, col, multivalued=sd.multivalued)
                logger.debug(f"Facet query: {facet_query_str}")
                rows = list(conn.execute(text(facet_query_str)))
                results[col] = [tuple(row) for row in rows]
            return results

    def _sqla_table(self, cd: ClassDefinition) -> Table:
        schema_view = self.parent.schema_view
        metadata_obj = sqla.MetaData()
        cols = []
        for att in schema_view.class_induced_slots(cd.name):
            typ = TMAP.get(att.range, sqla.String)
            if att.inlined or att.inlined_as_list:
                typ = sqla.JSON
            if att.multivalued:
                typ = sqla.ARRAY(typ, dimensions=1)
            if att.array:
                typ = sqla.ARRAY(typ, dimensions=1)
            col = Column(att.name, typ)
            cols.append(col)
        t = Table(self.alias, metadata_obj, *cols)
        return t

    def _check_if_initialized(self) -> bool:
        # if self._initialized:
        #    return True
        query = Query(
            from_table="information_schema.tables", where_clause={"table_type": "BASE TABLE", "table_name": self.alias}
        )
        qr = self.parent.query(query)
        if qr.num_rows > 0:
            return True
        return False

    def _create_table(self, cd: ClassDefinition):
        if self._table_created or self.metadata.is_prepopulated:
            logger.info(f"Already have table for: {cd.name}")
            return
        query = Query(
            from_table="information_schema.tables", where_clause={"table_type": "BASE TABLE", "table_name": self.alias}
        )
        qr = self.parent.query(query)
        if qr.num_rows > 0:
            logger.info(f"Table already exists for {cd.name}")
            self._table_created = True
            self._initialized = True
            self.metadata.is_prepopulated = True
            return
        logger.info(f"Creating table for {cd.name}")
        t = self._sqla_table(cd)
        ct = CreateTable(t)
        ddl = str(ct.compile(self.parent.engine))
        with self.parent.engine.connect() as conn:
            conn.execute(text(ddl))
            conn.commit()
        self._table_created = True
        self._initialized = True
        self.metadata.is_prepopulated = True



import logging
from pathlib import Path
from typing import Optional

import yaml
from linkml.utils.schema_builder import SchemaBuilder
from linkml_runtime import SchemaView

from linkml_store.api import Database
from linkml_store.api.config import DatabaseConfig
from linkml_store.api.stores.filesystem.filesystem_collection import FileSystemCollection
from linkml_store.utils.file_utils import safe_remove_directory
from linkml_store.utils.format_utils import Format, load_objects

logger = logging.getLogger(__name__)


class FileSystemDatabase(Database):
    collection_class = FileSystemCollection

    directory_path: Optional[Path] = None
    default_file_format: Optional[str] = None

    no_backup_on_drop: bool = False

    def __init__(self, handle: Optional[str] = None, **kwargs):
        handle = handle.replace("file:", "")
        if handle.startswith("//"):
            handle = handle[2:]
        self.directory_path = Path(handle)
        self.load_metadata()
        super().__init__(handle=handle, **kwargs)

    @property
    def metadata_path(self) -> Path:
        return self.directory_path / ".linkml_metadata.yaml"

    def load_metadata(self):
        if self.metadata_path.exists():
            md_dict = yaml.safe_load(open(self.metadata_path))
            metadata = DatabaseConfig(**md_dict)
        else:
            metadata = DatabaseConfig()
        self.metadata = metadata

    def close(self, **kwargs):
        pass

    def drop(self, no_backup=False, **kwargs):
        self.close()
        path = self.directory_path
        if path.exists():
            safe_remove_directory(path, no_backup=self.no_backup_on_drop or no_backup)

    def init_collections(self):
        metadata = self.metadata
        if self._collections is None:
            self._collections = {}
        for name, collection_config in metadata.collections.items():
            collection = FileSystemCollection(parent=self, **collection_config.dict())
            self._collections[name] = collection
        path = self.directory_path
        if path.exists():
            for fmt in Format:
                suffix = fmt.value
                logger.info(f"Looking for {suffix} files in {path}")
                for f in path.glob(f"*.{suffix}"):
                    logger.info(f"Found {f}")
                    n = f.stem
                    objs = load_objects(f, suffix, expected_type=list)
                    collection = FileSystemCollection(parent=self, name=n)
                    self._collections[n] = collection
                    collection._set_objects(objs)

    def xxxinduce_schema_view(self) -> SchemaView:
        logger.info(f"Inducing schema view for {self.handle}")
        sb = SchemaBuilder()

        for collection_name in self.list_collection_names():
            sb.add_class(collection_name)
        return SchemaView(sb.schema)



from collections import namedtuple
from typing import Any, Dict, List, Optional, Tuple, Union

import pandas as pd
from pydantic import BaseModel

# defined a named tuple called between with two values (start, end):
# This is used in the Query class to represent a range of values
# This is used in the Query class to represent a range of values
Between = namedtuple("Between", "min max")

FACET_GROUP_ATOM = Union[str, int, float, Between]
FACET_GROUP = Union[FACET_GROUP_ATOM, Tuple[FACET_GROUP_ATOM, ...]]


class Query(BaseModel):
    """
    A query object.

    - In SQL this would be a SQL query string
    """

    from_table: Optional[str]
    select_cols: Optional[List[str]] = None
    where_clause: Optional[Union[str, List[str], Dict[str, Any]]] = None
    sort_by: Optional[List[str]] = None
    limit: Optional[int] = None
    offset: Optional[int] = None
    include_facet_counts: bool = False
    facet_slots: Optional[List[str]] = None


class FacetCountResult(BaseModel):
    """
    A facet count result
    """

    as_dict: Dict[FACET_GROUP, List[Tuple[FACET_GROUP, int]]]


class QueryResult(BaseModel):
    """
    A query result
    """

    query: Optional[Query] = None
    search_term: Optional[str] = None
    num_rows: int
    offset: Optional[int] = 0
    rows: Optional[List[Dict[str, Any]]] = None
    ranked_rows: Optional[List[Tuple[float, Dict[str, Any]]]] = None
    _rows_dataframe: Optional[pd.DataFrame] = None
    facet_counts: Optional[Dict[str, List[Tuple[FACET_GROUP, int]]]] = None

    @property
    def rows_dataframe(self) -> pd.DataFrame:
        if self.ranked_rows is not None:
            self._rows_dataframe = pd.DataFrame([{"score": score, **row} for score, row in self.ranked_rows])
        if self._rows_dataframe is None and self.rows:
            self._rows_dataframe = pd.DataFrame(self.rows)
        return self._rows_dataframe

    def set_rows(self, rows: pd.DataFrame):
        self._rows_dataframe = rows

    class Config:
        arbitrary_types_allowed = True



from typing import Dict, List, Tuple, Union

import pandas as pd


def facet_summary_to_dataframe_unmelted(
    facet_summary: Dict[Union[str, Tuple[str, ...]], List[Tuple[Union[str, Tuple[str, ...]], int]]]
) -> pd.DataFrame:
    rows = []

    for facet_type, facet_data in facet_summary.items():
        if isinstance(facet_type, str):
            # Single facet type
            for category, value in facet_data:
                rows.append({facet_type: category, "Value": value})
        else:
            # Multiple facet types
            for cat_val_tuple in facet_data:
                if len(cat_val_tuple) == 2:
                    categories, value = cat_val_tuple
                else:
                    categories, value = cat_val_tuple[:-1], cat_val_tuple[-1]
                row = {"Value": value}
                for i, facet in enumerate(facet_type):
                    row[facet] = categories[i]
                rows.append(row)

    df = pd.DataFrame(rows)

    # Ensure all columns are present, fill with None if missing
    all_columns = set(col for facet in facet_summary.keys() for col in (facet if isinstance(facet, tuple) else [facet]))
    for col in all_columns:
        if col not in df.columns:
            df[col] = None

    # Move 'Value' to the end
    cols = [col for col in df.columns if col != "Value"] + ["Value"]
    df = df[cols]

    return df



# flake8: noqa: E402
from linkml_store.api.collection import Collection
from linkml_store.api.database import Database
from linkml_store.api.client import Client

# flake8: noqa

__all__ = ["Client", "Database", "Collection"]



import json
import logging
from typing import Any, Dict, List, Optional, Tuple, Union

import h5py

from linkml_store.api import Collection
from linkml_store.api.collection import DEFAULT_FACET_LIMIT, OBJECT
from linkml_store.api.queries import Query, QueryResult

logger = logging.getLogger(__name__)


class HDF5Collection(Collection):

    @property
    def hdf5_group(self) -> h5py.Group:
        return self.parent.file[self.name]

    def insert(self, objs: Union[OBJECT, List[OBJECT]], **kwargs):
        if not isinstance(objs, list):
            objs = [objs]

        for obj in objs:
            if "id" not in obj:
                raise ValueError("Each object must have an 'id' field.")
            obj_id = str(obj["id"])
            for key, value in obj.items():
                if key == "id":
                    continue
                if isinstance(value, (dict, list)):
                    value = json.dumps(value)
                self.hdf5_group.create_dataset(f"{obj_id}/{key}", data=value)

    def delete(self, objs: Union[OBJECT, List[OBJECT]], **kwargs) -> int:
        if not isinstance(objs, list):
            objs = [objs]
        count = 0
        for obj in objs:
            if "id" not in obj:
                raise ValueError("Each object must have an 'id' field.")
            obj_id = str(obj["id"])
            if obj_id in self.hdf5_group:
                del self.hdf5_group[obj_id]
                count += 1
        return count

    def delete_where(self, where: Optional[Dict[str, Any]] = None, missing_ok=True, **kwargs) -> int:
        logger.info(f"Deleting from {self.target_class_name} where: {where}")
        if where is None:
            where = {}
        results = self.query(Query(where_clause=where)).rows
        count = self.delete(results)
        return count

    def query(self, query: Query, **kwargs) -> QueryResult:
        results = []
        for obj_id in self.hdf5_group:
            obj = {"id": obj_id}
            for key, value in self.hdf5_group[obj_id].items():
                try:
                    obj[key] = json.loads(value[()])
                except json.JSONDecodeError:
                    obj[key] = value[()]
            if self._match_where_clause(obj, query.where_clause):
                results.append(obj)

        count = len(results)
        if query.limit:
            results = results[: query.limit]
        return QueryResult(query=query, num_rows=count, rows=results)

    def query_facets(
        self, where: Dict = None, facet_columns: List[str] = None, facet_limit=DEFAULT_FACET_LIMIT, **kwargs
    ) -> Dict[str, List[Tuple[Any, int]]]:
        results = {}
        if not facet_columns:
            facet_columns = list(self.class_definition().attributes.keys())

        for col in facet_columns:
            logger.debug(f"Faceting on {col}")
            facet_counts = {}
            for obj in self.query(Query(where_clause=where)).rows:
                if col in obj:
                    value = obj[col]
                    if isinstance(value, list):
                        for v in value:
                            facet_counts[v] = facet_counts.get(v, 0) + 1
                    else:
                        facet_counts[value] = facet_counts.get(value, 0) + 1
            facet_counts = sorted(facet_counts.items(), key=lambda x: x[1], reverse=True)[:facet_limit]
            results[col] = facet_counts

        return results

    def _match_where_clause(self, obj: Dict[str, Any], where_clause: Optional[Dict[str, Any]]) -> bool:
        if where_clause is None:
            return True
        for key, value in where_clause.items():
            if key not in obj:
                return False
            if obj[key] != value:
                return False
        return True



import logging
import os
import subprocess
from pathlib import Path
from typing import Optional
from urllib.parse import urlparse

from pymongo import MongoClient
from pymongo.database import Database

logger = logging.getLogger(__name__)


def extract_connection_info(db: Database):
    client = db.client

    # Get the host and port
    host_info = client.address
    if host_info:
        host, port = host_info
    else:
        # For replica sets or sharded clusters, we might need to get this differently
        host = client.HOST
        port = client.PORT

    # Get the database name
    db_name = db.name

    # Get username if available
    username = None
    if hasattr(client, "options") and hasattr(client.options, "credentials"):
        credentials = client.options.credentials
        if credentials:
            username = credentials.username

    return {"host": host, "port": port, "db_name": db_name, "username": username}


def get_connection_string(client: MongoClient):
    """
    Extract a connection string from the MongoClient.
    This avoids triggering truth value testing on Database objects.
    """
    if client.address:
        host, port = client.address
        return f"{host}:{port}"
    if hasattr(client, "address") and client.address:
        host, port = client.address
        return f"{host}:{port}"
    elif client.hosts:
        # For replica sets, return all hosts
        return ",".join(f"{host}:{port}" for host, port in client.hosts)
    elif hasattr(client, "HOST"):
        # If we can't determine hosts, use the entire URI
        parsed_uri = urlparse(client.HOST)
        return f"{parsed_uri.hostname}:{parsed_uri.port}"
    else:
        raise ValueError("Unable to determine connection string from client")


def get_connection_info(db: Database):
    """
    Extract connection information from the Database object.
    """
    # Get the name of the database
    db_name = db.name

    # Get the client's node list (this should work for single nodes and replica sets)
    node_list = db.client.nodes

    if not node_list:
        raise ValueError("Unable to determine connection information from database")

    # Use the first node in the list (for single node setups, this will be the only node)
    first_node = node_list[0]
    host, port = first_node

    return host, port, db_name


def get_auth_from_client(client: MongoClient):
    """Extract authentication details from MongoClient."""
    if hasattr(client, "_MongoClient__options"):
        # For older versions of PyMongo
        options = client._MongoClient__options
    elif hasattr(client, "options"):
        # For newer versions of PyMongo
        options = client.options
    else:
        return None, None, None

    if hasattr(options, "credentials"):
        creds = options.credentials
        return creds.username, creds.password, creds.source
    return None, None, None


def connection_from_handle(handle: str):
    if handle.startswith("mongodb://"):
        handle = handle.replace("mongodb://", "")
    host, db = handle.split("/")
    return host, db


def export_mongodb(handle: str, location: str, password: Optional[str] = None):
    host, db_name = connection_from_handle(handle)

    # Construct the mongodump command
    cmd = ["mongodump", f"--host={host}", f"--db={db_name}"]
    logger.info(f"Exporting MongoDB database {db_name} from {host} to {location}")
    cmd.extend(["--out", location])
    result = subprocess.run(cmd, check=True, capture_output=True, text=True)
    logger.info(f"MongoDB export completed successfully. Output: {result.stdout}")


def import_mongodb(handle: str, dump_dir: str, drop: bool = False):
    host, db_name = connection_from_handle(handle)

    # list dirs in dump_dir
    dir_path = Path(dump_dir)
    if not dir_path.is_dir():
        raise ValueError(f"{dir_path} is not a dir")
    directories = [name for name in os.listdir(dump_dir)]
    if len(directories) != 1:
        raise ValueError(f"Expected exactly one database in {dump_dir}, got: {directories}")
    src_db_name = directories[0]

    # Construct the mongorestore command
    cmd = [
        "mongorestore",
        f"--host={host}",
        f"--nsFrom={src_db_name}.*",
        f"--nsTo={db_name}.*",
        str(dump_dir),
    ]

    # Add drop option if specified
    if drop:
        cmd.append("--drop")
    logger.info(f"CMD={cmd}")
    # Execute mongorestore
    result = subprocess.run(cmd, check=True, capture_output=True, text=True)
    if result.stderr:
        logger.warning(result.stderr)
    logger.info(f"MongoDB import completed successfully. Output: {result.stdout} // {result.stderr}")



import logging
from abc import ABC
from collections import defaultdict
from copy import copy
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    Generic,
    Iterator,
    List,
    Optional,
    Sequence,
    Type,
    Union,
)

from linkml_store.api.types import CollectionType
from linkml_store.utils.format_utils import Format, load_objects, render_output
from linkml_store.utils.patch_utils import PatchDict

try:
    from linkml.validator.report import Severity, ValidationResult
except ImportError:
    ValidationResult = None

from linkml_runtime import SchemaView
from linkml_runtime.linkml_model import ClassDefinition, SchemaDefinition

from linkml_store.api.collection import Collection
from linkml_store.api.config import CollectionConfig, DatabaseConfig
from linkml_store.api.queries import Query, QueryResult

if TYPE_CHECKING:
    from linkml_store.api.client import Client

logger = logging.getLogger(__name__)

LISTENER = Callable[[Collection, List[PatchDict]], None]


class Database(ABC, Generic[CollectionType]):
    """
    A Database provides access to named collections of data.

    A database object is owned by a :ref:`Client`. The database
    object uses a :ref:`handle` to know what kind of external
    dataase system to connect to (e.g. duckdb, mongodb). The handle
    is a string ``<DatabaseType>:<LocalLocator>``

    The
    database object may also have an :ref:`alias` that is mapped
    to the handle.

    Attaching a database
    --------------------
    >>> from linkml_store.api.client import Client
    >>> client = Client()
    >>> db = client.attach_database("duckdb:///:memory:", alias="test")

    We can check the value of the handle:

    >>> db.handle
    'duckdb:///:memory:'

    The alias can be used to retrieve the database object from the client

    >>> assert db == client.get_database("test")

    Creating a collection
    ---------------------

    >>> collection = db.create_collection("Person")
    >>> len(db.list_collections())
    1
    >>> db.get_collection("Person") == collection
    True
    >>> objs = [{"id": "P1", "name": "John", "age_in_years": 30}, {"id": "P2", "name": "Alice", "age_in_years": 25}]
    >>> collection.insert(objs)
    >>> qr = collection.find()
    >>> len(qr.rows)
    2
    >>> qr.rows[0]["id"]
    'P1'
    >>> qr.rows[1]["name"]
    'Alice'
    >>> qr = collection.find({"name": "John"})
    >>> len(qr.rows)
    1
    >>> qr.rows[0]["name"]
    'John'

    """

    _schema_view: Optional[SchemaView] = None
    """Schema for the database. May be transformed."""

    _original_schema_view: Optional[SchemaView] = None
    """If a schema must be transformed, then the original is stored here."""

    _collections: Optional[Dict[str, Collection]] = None
    parent: Optional["Client"] = None
    metadata: Optional[DatabaseConfig] = None
    collection_class: ClassVar[Optional[Type[Collection]]] = None

    listeners: Optional[List[LISTENER]] = None

    def __init__(self, handle: Optional[str] = None, metadata: Optional[DatabaseConfig] = None, **kwargs):
        if metadata:
            self.metadata = metadata
        else:
            self.metadata = DatabaseConfig(handle=handle, **kwargs)
        if handle is not None and self.metadata.handle is not None and handle != self.metadata.handle:
            raise ValueError(f"Handle mismatch: {handle} != {self.metadata.handle}")
        self._initialize_schema()
        self._initialize_collections()

    def _initialize_schema(self, **kwargs):
        db_config = self.metadata
        if db_config.schema_location:
            schema_location = db_config.schema_location.format(base_dir=self.parent.metadata.base_dir)
            logger.info(f"Loading schema from: {schema_location}")
            self.load_schema_view(schema_location)
        if db_config.schema_dict:
            schema_dict = copy(db_config.schema_dict)
            if "id" not in schema_dict:
                schema_dict["id"] = "tmp"
            if "name" not in schema_dict:
                schema_dict["name"] = "tmp"
            self.set_schema_view(SchemaView(SchemaDefinition(**schema_dict)))

    def from_config(self, db_config: DatabaseConfig, **kwargs):
        """
        Initialize a database from a configuration.

        TODO: DEPRECATE

        :param db_config: database configuration
        :param kwargs: additional arguments
        """
        self.metadata = db_config
        self._initialize_schema()
        self._initialize_collections()
        return self

    def _initialize_collections(self):
        if not self.metadata.collections:
            return
        for k, collection_config in self.metadata.collections.items():
            if collection_config.alias:
                if collection_config.alias != k:
                    raise ValueError(f"Alias mismatch: {collection_config.alias} != {k}")
            alias = k
            typ = collection_config.type or alias
            _collection = self.create_collection(typ, alias=alias, metadata=collection_config)
            assert _collection.alias == alias
            assert _collection.target_class_name == typ
            if collection_config.attributes:
                # initialize schema
                sv = self.schema_view
                cd = ClassDefinition(typ, attributes=collection_config.attributes)
                sv.schema.classes[cd.name] = cd
                sv.set_modified()
                # assert collection.class_definition() is not None

    @property
    def recreate_if_exists(self) -> bool:
        """
        Return whether to recreate the database if it already exists.

        :return:
        """
        return self.metadata.recreate_if_exists

    @property
    def handle(self) -> str:
        """
        Return the database handle.

        Examples:

        - ``duckdb:///:memory:``
        - ``duckdb:///tmp/test.db``
        - ``mongodb://localhost:27017/``

        :return:
        """
        return self.metadata.handle

    @property
    def alias(self):
        return self.metadata.alias

    def store(self, obj: Dict[str, Any], **kwargs):
        """
        Store an object in the database.

        The object is assumed to be a Dictionary of Collections.

        >>> from linkml_store.api.client import Client
        >>> client = Client()
        >>> db = client.attach_database("duckdb", alias="test")
        >>> db.store({"persons": [{"id": "P1", "name": "John", "age_in_years": 30}]})
        >>> collection = db.get_collection("persons")
        >>> qr = collection.find()
        >>> qr.num_rows
        1

        :param obj: object to store
        :param kwargs: additional arguments
        """
        sv = self.schema_view
        roots = [c for c in sv.all_classes().values() if c.tree_root]
        root = roots[0] if roots else None
        for k, v in obj.items():
            logger.info(f"Storing collection {k}")
            if root:
                slot = sv.induced_slot(k, root.name)
                if not slot:
                    raise ValueError(f"Cannot determine type for {k}")
            else:
                slot = None
            if isinstance(v, dict):
                logger.debug(f"Coercing dict to list: {v}")
                v = [v]
            if not isinstance(v, list):
                continue
            if not v:
                continue
            if slot:
                logger.debug(f"Aligning to existing slot: {slot.name} range={slot.range}")
                collection = self.get_collection(slot.name, type=slot.range, create_if_not_exists=True)
            else:
                collection = self.get_collection(k, create_if_not_exists=True)
            logger.debug(f"Replacing using {collection.alias} {collection.target_class_name}")
            collection.replace(v)

    def commit(self, **kwargs):
        """
        Commit pending changes to the database.

        :param kwargs:
        :return:
        """
        for coll in self.list_collections():
            coll.commit()

    def close(self, **kwargs):
        """
        Close the database.

        :param kwargs:
        :return:
        """
        raise NotImplementedError()

    @property
    def _collection_class(self) -> Type[Collection]:
        raise NotImplementedError()

    def create_collection(
        self,
        name: str,
        alias: Optional[str] = None,
        metadata: Optional[CollectionConfig] = None,
        recreate_if_exists=False,
        **kwargs,
    ) -> CollectionType:
        """
        Create a new collection in the current database.

        The collection must have a *Type*, and may have an *Alias*.

        Examples:

        >>> from linkml_store.api.client import Client
        >>> client = Client()
        >>> db = client.attach_database("duckdb", alias="test")
        >>> collection = db.create_collection("Person", alias="persons")
        >>> collection.alias
        'persons'
        >>> collection.target_class_name
        'Person'

        If alias is not provided, it defaults to the name of the type.

        >>> collection = db.create_collection("Organization")
        >>> collection.alias
        'Organization'

        :param name: name of the collection
        :param alias: alias for the collection
        :param metadata: metadata for the collection
        :param recreate_if_exists: recreate the collection if it already exists
        :param kwargs: additional arguments
        """
        if not name:
            raise ValueError(f"Collection name must be provided: alias: {alias} metadata: {metadata}")
        collection_cls = self.collection_class
        collection = collection_cls(name=name, parent=self, metadata=metadata)
        if alias:
            collection.metadata.alias = alias
        if metadata and metadata.source:
            collection.load_from_source()
        if metadata and metadata.attributes:
            sv = self.schema_view
            schema = sv.schema
            cd = ClassDefinition(name=metadata.type, attributes=metadata.attributes)
            schema.classes[cd.name] = cd
        if not self._collections:
            self._collections = {}
        if not alias:
            alias = name
        self._collections[alias] = collection
        if recreate_if_exists:
            logger.debug(f"Recreating collection {collection.alias}")
            collection.delete_where({}, missing_ok=True)
        return collection

    def list_collections(self, include_internal=False) -> Sequence[Collection]:
        """
        List all collections.

        Examples
        --------
        >>> from linkml_store.api.client import Client
        >>> client = Client()
        >>> db = client.attach_database("duckdb", alias="test")
        >>> c1 = db.create_collection("Person")
        >>> c2 = db.create_collection("Product")
        >>> collections = db.list_collections()
        >>> len(collections)
        2
        >>> [c.target_class_name for c in collections]
        ['Person', 'Product']

        :param include_internal: include internal collections
        :return: list of collections
        """
        if not self._collections:
            self.init_collections()
        return [c for c in self._collections.values() if include_internal or not c.is_internal]

    def list_collection_names(self, **kwargs) -> Sequence[str]:
        """
        List all collection names.

        Examples
        --------
        >>> from linkml_store.api.client import Client
        >>> client = Client()
        >>> db = client.attach_database("duckdb", alias="test")
        >>> c1 = db.create_collection("Person")
        >>> c2 = db.create_collection("Product")
        >>> collection_names = db.list_collection_names()
        >>> len(collection_names)
        2
        >>> collection_names
        ['Person', 'Product']

        """
        return [c.alias for c in self.list_collections(**kwargs)]

    def get_collection(
        self, name: str, type: Optional[str] = None, create_if_not_exists=True, **kwargs
    ) -> "Collection":
        """
        Get a named collection.

        Examples
        --------
        >>> from linkml_store.api.client import Client
        >>> client = Client()
        >>> db = client.attach_database("duckdb", alias="test")
        >>> collection = db.create_collection("Person")
        >>> db.get_collection("Person") == collection
        True
        >>> db.get_collection("NonExistent", create_if_not_exists=False)
        Traceback (most recent call last):
            ...
        KeyError: 'Collection NonExistent does not exist'

        :param name: name of the collection
        :param type: target class name
        :param create_if_not_exists: create the collection if it does not exist

        """
        if not self._collections:
            logger.debug("Initializing collections")
            self.init_collections()
        if name not in self._collections.keys():
            if create_if_not_exists:
                if type is None:
                    type = name
                logger.debug(f"Creating new collection: {name} kwargs: {kwargs}")
                self._collections[name] = self.create_collection(type, alias=name, **kwargs)
            else:
                raise KeyError(f"Collection {name} does not exist")
        return self._collections[name]

    def init_collections(self):
        """
        Initialize collections.

        TODO: Not typically called directly: consider making this private
        :return:
        """
        raise NotImplementedError

    def query(self, query: Query, **kwargs) -> QueryResult:
        """
        Run a query against the database.

        Examples
        --------
        >>> from linkml_store.api.client import Client
        >>> from linkml_store.api.queries import Query
        >>> client = Client()
        >>> db = client.attach_database("duckdb", alias="test")
        >>> collection = db.create_collection("Person")
        >>> collection.insert([{"id": "P1", "name": "John"}, {"id": "P2", "name": "Alice"}])
        >>> query = Query(from_table="Person", where_clause={"name": "John"})
        >>> result = db.query(query)
        >>> len(result.rows)
        1
        >>> result.rows[0]["id"]
        'P1'

        :param query:
        :param kwargs:
        :return:

        """
        if query.from_table:
            collection = self.get_collection(query.from_table)
            return collection.query(query, **kwargs)
        else:
            raise NotImplementedError(f"Querying without a table is not supported in {self.__class__.__name__}")

    @property
    def schema_view(self) -> SchemaView:
        """
        Return a schema view for the named collection.

        If no explicit schema is provided, this will generalize one

        Induced schema example:

        >>> from linkml_store.api.client import Client
        >>> client = Client()
        >>> db = client.attach_database("duckdb", alias="test")
        >>> collection = db.create_collection("Person", alias="persons")
        >>> collection.insert([{"id": "P1", "name": "John", "age_in_years": 25}])
        >>> schema_view = db.schema_view
        >>> cd = schema_view.get_class("Person")
        >>> cd.attributes["id"].range
        'string'
        >>> cd.attributes["age_in_years"].range
        'integer'

        We can reuse the same class:

        >>> collection2 = db.create_collection("Person", alias="other_persons")
        >>> collection2.class_definition().attributes["age_in_years"].range
        'integer'
        """
        if not self._schema_view:
            self._initialize_schema()
        if not self._schema_view:
            self._schema_view = self.induce_schema_view()
        return self._schema_view

    def set_schema_view(self, schema_view: Union[str, Path, SchemaView]):
        """
        Set the schema view for the database.

        >>> from linkml_store.api.client import Client
        >>> client = Client()
        >>> db = client.attach_database("duckdb", alias="test")
        >>> sv = SchemaView("tests/input/countries/countries.linkml.yaml")
        >>> db.set_schema_view(sv)
        >>> cd = db.schema_view.schema.classes["Country"]
        >>> sorted(cd.slots)
        ['capital', 'code', 'continent', 'languages', 'name']
        >>> induced_slots = {s.name: s for s in sv.class_induced_slots("Country")}
        >>> sorted(induced_slots.keys())
        ['capital', 'code', 'continent', 'languages', 'name']
        >>> induced_slots["code"].identifier
        True

        Creating a new collection will align with the schema view:

        >>> collection = db.create_collection("Country", "all_countries")
        >>> sorted(collection.class_definition().slots)
        ['capital', 'code', 'continent', 'languages', 'name']

        :param schema_view: can be either a path to the schema, or a SchemaView object
        :return:
        """
        if isinstance(schema_view, Path):
            schema_view = str(schema_view)
        if isinstance(schema_view, str):
            schema_view = SchemaView(schema_view)
        self._schema_view = schema_view
        # self._schema_view = SchemaView(schema_view.materialize_derived_schema())
        if not self._collections:
            return

        # align with induced schema
        roots = [c for c in schema_view.all_classes().values() if c.tree_root]
        if len(roots) == 0:
            all_ranges = set()
            for cn in schema_view.all_classes():
                for slot in schema_view.class_induced_slots(cn):
                    if slot.range:
                        all_ranges.add(slot.range)
            roots = [
                c
                for c in schema_view.all_classes().values()
                if not all_ranges.intersection(schema_view.class_ancestors(c.name, reflexive=True))
            ]
        if len(roots) == 1:
            root = roots[0]
            for slot in schema_view.class_induced_slots(root.name):
                inlined = slot.inlined or slot.inlined_as_list
                if inlined and slot.range:
                    if slot.name in self._collections:
                        coll = self._collections[slot.name]
                        coll.metadata.type = slot.range

    def load_schema_view(self, path: Union[str, Path]):
        """
        Load a schema view from a file.

        >>> from linkml_store.api.client import Client
        >>> client = Client()
        >>> db = client.attach_database("duckdb", alias="test")
        >>> db.load_schema_view("tests/input/countries/countries.linkml.yaml")
        >>> sv = db.schema_view
        >>> cd = sv.schema.classes["Country"]
        >>> sorted(cd.slots)
        ['capital', 'code', 'continent', 'languages', 'name']
        >>> induced_slots = {s.name: s for s in sv.class_induced_slots("Country")}
        >>> sorted(induced_slots.keys())
        ['capital', 'code', 'continent', 'languages', 'name']
        >>> induced_slots["code"].identifier
        True

        Creating a new collection will align with the schema view:

        >>> collection = db.create_collection("Country", "all_countries")
        >>> sorted(collection.class_definition().slots)
        ['capital', 'code', 'continent', 'languages', 'name']

        :param path:
        :return:
        """
        if isinstance(path, Path):
            path = str(path)
        self.set_schema_view(SchemaView(path))

    def induce_schema_view(self) -> SchemaView:
        """
        Induce a schema view from a schema definition.

        >>> from linkml_store.api.client import Client
        >>> from linkml_store.api.queries import Query
        >>> client = Client()
        >>> db = client.attach_database("duckdb", alias="test")
        >>> collection = db.create_collection("Person")
        >>> collection.insert([{"id": "P1", "name": "John", "age_in_years": 25},
        ...                 {"id": "P2", "name": "Alice", "age_in_years": 25}])
        >>> schema_view = db.induce_schema_view()
        >>> cd = schema_view.get_class("Person")
        >>> cd.attributes["id"].range
        'string'
        >>> cd.attributes["age_in_years"].range
        'integer'

        :return: A schema view
        """
        logger.info(f"Inducing schema view for {self.handle}")
        from linkml_runtime.utils.schema_builder import SchemaBuilder

        sb = SchemaBuilder()

        for collection_name in self.list_collection_names():
            coll = self.get_collection(collection_name)
            sb.add_class(coll.target_class_name)
        return SchemaView(sb.schema)

    def iter_validate_database(self, **kwargs) -> Iterator["ValidationResult"]:
        """
        Validate the contents of the database.

        An an example, let's create a database with a predefined schema
        from the countries.linkml.yaml file:

        >>> from linkml_store.api.client import Client
        >>> client = Client()
        >>> db = client.attach_database("duckdb", alias="test")
        >>> db.load_schema_view("tests/input/countries/countries.linkml.yaml")

        Let's introspect the schema to see what slots are applicable for the class "Country":

        >>> sv = db.schema_view
        >>> for slot in sv.class_induced_slots("Country"):
        ...     print(slot.name, slot.range, slot.required)
        name string True
        code string True
        capital string True
        continent string True
        languages Language None

        Next we'll create a collection, binding it to the target class "Country", and insert
        valid data:

        >>> collection = db.create_collection("Country", "all_countries")
        >>> obj = {"code": "US", "name": "United States", "continent": "North America", "capital": "Washington, D.C."}
        >>> collection.insert([obj])
        >>> list(db.iter_validate_database())
        []

        Now let's insert some invalid data (missing required fields)

        >>> collection.insert([{"code": "FR", "name": "France"}])
        >>> for r in db.iter_validate_database():
        ...    print(r.message[0:32])
        'capital' is a required property
        'continent' is a required proper

        :param kwargs:
        :return: iterator over validation results
        """
        for collection in self.list_collections():
            yield from collection.iter_validate_collection(**kwargs)
        if self.metadata.ensure_referential_integrity:
            yield from self._validate_referential_integrity(**kwargs)

    def _validate_referential_integrity(self, **kwargs) -> Iterator["ValidationResult"]:
        """
        Validate referential integrity of the database.

        :param kwargs:
        :return: iterator over validation results
        """
        sv = self.schema_view
        cmap = defaultdict(list)
        for collection in self.list_collections():
            if not collection.target_class_name:
                raise ValueError(f"Collection {collection.name} has no target class")
            cmap[collection.target_class_name].append(collection)
        for collection in self.list_collections():
            cd = collection.class_definition()
            induced_slots = sv.class_induced_slots(cd.name)
            slot_map = {s.name: s for s in induced_slots}
            # rmap = {s.name: s.range for s in induced_slots}
            sr_to_coll = {s.name: cmap.get(s.range, []) for s in induced_slots if s.range}
            for obj in collection.find_iter():
                for k, v in obj.items():
                    if k not in sr_to_coll:
                        continue
                    ref_colls = sr_to_coll[k]
                    if not ref_colls:
                        continue
                    if not isinstance(v, (str, int)):
                        continue
                    slot = slot_map[k]
                    found = False
                    for ref_coll in ref_colls:
                        ref_obj = ref_coll.get_one(v)
                        if ref_obj:
                            found = True
                            break
                    if not found:
                        yield ValidationResult(
                            type="ReferentialIntegrity",
                            severity=Severity.ERROR,
                            message=f"Referential integrity error: {slot.range} not found",
                            instantiates=slot.range,
                            instance=v,
                        )

    def drop(self, **kwargs):
        """
        Drop the database and all collections.

        >>> from linkml_store.api.client import Client
        >>> client = Client()
        >>> path = Path("/tmp/test.db")
        >>> path.parent.mkdir(exist_ok=True, parents=True)
        >>> db = client.attach_database(f"duckdb:///{path}")
        >>> db.store({"persons": [{"id": "P1", "name": "John", "age_in_years": 30}]})
        >>> coll = db.get_collection("persons")
        >>> coll.find({}).num_rows
        1
        >>> db.drop()
        >>> db = client.attach_database("duckdb:///tmp/test.db", alias="test")
        >>> coll = db.get_collection("persons")
        >>> coll.find({}).num_rows
        0

        :param kwargs: additional arguments
        """
        raise NotImplementedError()

    def import_database(self, location: str, source_format: Optional[Union[str, Format]] = None, **kwargs):
        """
        Import a database from a file or location.

        :param location: location of the file
        :param source_format: source format
        :param kwargs: additional arguments
        """
        if isinstance(source_format, str):
            source_format = Format(source_format)
        if isinstance(source_format, Format):
            if source_format.is_dump_format() and source_format in [Format.SQLDUMP_DUCKDB, Format.DUMP_MONGODB]:
                # import into a test instance
                tmp_handle = source_format.value
                client = self.parent
                tmp_db = client.attach_database(tmp_handle, alias="tmp")
                # TODO: check for infinite recursion
                tmp_db.import_database(location, source_format=source_format)
                obj = {}
                for coll in tmp_db.list_collections():
                    qr = coll.find({}, limit=-1)
                    obj[coll.alias] = qr.rows
                self.store(obj)
                return
        objects = load_objects(location, format=source_format)
        for obj in objects:
            self.store(obj)

    def export_database(self, location: str, target_format: Optional[Union[str, Format]] = None, **kwargs):
        """
        Export a database to a file or location.

        :param location: location of the file
        :param target_format: target format
        :param kwargs: additional arguments
        """
        obj = {}
        if isinstance(target_format, str):
            target_format = Format(target_format)
        for coll in self.list_collections():
            qr = coll.find({}, limit=-1)
            obj[coll.alias] = qr.rows
        logger.info(f"Exporting object with {len(obj)} collections to {location} in {target_format} format")
        if isinstance(target_format, Format):
            if target_format.is_dump_format() and target_format in [Format.SQLDUMP_DUCKDB, Format.DUMP_MONGODB]:
                tmp_handle = target_format.value
                client = self.parent
                tmp_db = client.attach_database(tmp_handle, alias="tmp")
                tmp_db.store(obj)
                # TODO: check for infinite recursion
                tmp_db.export_database(location, target_format=target_format)
                return
        if Path(location).is_dir():
            raise ValueError(f"{location} is a directory; cannot write {target_format} to a dir")
        with open(location, "w", encoding="utf-8") as stream:
            stream.write(render_output(obj, format=target_format))

    def broadcast(self, source: Collection, patches: List[PatchDict]):
        if not self.listeners:
            return
        for listener in self.listeners:
            listener(source, patches)



import logging
from pathlib import Path
from typing import TYPE_CHECKING, List

import numpy as np

from linkml_store.api.config import CollectionConfig
from linkml_store.index.indexer import INDEX_ITEM, Indexer

if TYPE_CHECKING:
    import llm


logger = logging.getLogger(__name__)


class LLMIndexer(Indexer):
    """
    An indexer that wraps the llm library.

    This indexer is used to convert text to vectors using the llm library.

    >>> indexer = LLMIndexer(cached_embeddings_database="tests/input/llm_cache.db")
    >>> vector = indexer.text_to_vector("hello")
    """

    embedding_model_name: str = "ada-002"
    _embedding_model: "llm.EmbeddingModel" = None
    cached_embeddings_database: str = None
    cached_embeddings_collection: str = None
    cache_queries: bool = False

    @property
    def embedding_model(self):
        import llm

        if self._embedding_model is None:
            self._embedding_model = llm.get_embedding_model(self.embedding_model_name)
        return self._embedding_model

    def text_to_vector(self, text: str, cache: bool = None, **kwargs) -> INDEX_ITEM:
        """
        Convert a text to an indexable object

        >>> indexer = LLMIndexer(cached_embeddings_database="tests/input/llm_cache.db")
        >>> vector = indexer.text_to_vector("hello")

        :param text:
        :return:
        """
        return self.texts_to_vectors([text], cache=cache, **kwargs)[0]

    def texts_to_vectors(self, texts: List[str], cache: bool = None, **kwargs) -> List[INDEX_ITEM]:
        """
        Use LLM to embed

        >>> indexer = LLMIndexer(cached_embeddings_database="tests/input/llm_cache.db")
        >>> vectors = indexer.texts_to_vectors(["hello", "goodbye"])

        :param texts:
        :return:
        """
        logging.info(f"Converting {len(texts)} texts to vectors")
        model = self.embedding_model
        if self.cached_embeddings_database and (cache is None or cache or self.cache_queries):
            model_id = model.model_id
            if not model_id:
                raise ValueError("Model ID is required to cache embeddings")
            db_path = Path(self.cached_embeddings_database)
            coll_name = self.cached_embeddings_collection
            if not coll_name:
                coll_name = "all_embeddings"
            from linkml_store import Client

            embeddings_client = Client()
            config = CollectionConfig(
                alias=coll_name,
                type="Embeddings",
                attributes={
                    "text": {"range": "string"},
                    "model_id": {"range": "string"},
                    "embedding": {"range": "float", "array": {}},
                },
            )
            embeddings_db = embeddings_client.get_database(f"duckdb:///{db_path}")
            if coll_name in embeddings_db.list_collection_names():
                # Load existing collection and use its model
                embeddings_collection = embeddings_db.create_collection(coll_name, metadata=config)
            else:
                embeddings_collection = embeddings_db.create_collection(coll_name, metadata=config)
            texts = list(texts)
            embeddings = list([None] * len(texts))
            uncached_texts = []
            n = 0
            for i in range(len(texts)):
                # TODO: optimize this
                text = texts[i]
                logger.info(f"Looking for cached embedding for {text}")
                r = embeddings_collection.find({"text": text, "model_id": model_id})
                if r.num_rows:
                    embeddings[i] = r.rows[0]["embedding"]
                    n += 1
                    logger.info("Found")
                else:
                    uncached_texts.append((text, i))
                    logger.info("NOT Found")
            logger.info(f"Found {n} cached embeddings")
            if uncached_texts:
                logger.info(f"Embedding {len(uncached_texts)} uncached texts")
                uncached_texts, uncached_indices = zip(*uncached_texts)
                uncached_embeddings = list(model.embed_multi(uncached_texts))
                # TODO: combine into a single insert with multiple rows
                for i, index in enumerate(uncached_indices):
                    logger.debug(f"Indexing text at {i}")
                    embeddings[index] = uncached_embeddings[i]
                    embeddings_collection.insert(
                        {"text": uncached_texts[i], "embedding": embeddings[index], "model_id": model_id}
                    )
                embeddings_collection.commit()
        else:
            logger.info(f"Embedding {len(texts)} texts")
            embeddings = model.embed_multi(texts)
        return [np.array(v, dtype=float) for v in embeddings]



import json
from copy import deepcopy
from typing import Any, Dict, List, Union

from pydantic import BaseModel


def object_path_update(
    obj: Union[BaseModel, Dict[str, Any]], path: str, value: Any
) -> Union[BaseModel, Dict[str, Any]]:
    """
    Updates a nested object based on a path description and a value. The path to the
    desired field is given in dot and bracket notation (e.g., 'a[0].b.c[1]').

    :param obj: The dictionary object to be updated.
    :type obj: Dict[str, Any]
    :param path: The path string indicating where to place the value within the object.
    :type path: str
    :param value: The value to be set at the specified path.
    :type value: Any
    :return: None. This function modifies the object in-place.
    :rtype: None

    **Example**::

    >>> data = {}
    >>> object_path_update(data, 'persons[0].foo.bar', 1)
    {'persons': [{'foo': {'bar': 1}}]}
    """
    if isinstance(obj, BaseModel):
        typ = type(obj)
        obj = obj.dict()
        obj = object_path_update(obj, path, value)
        return typ(**obj)
    obj = deepcopy(obj)
    ret_obj = obj
    parts = path.split(".")
    for part in parts[:-1]:
        if "[" in part:
            key, index = part[:-1].split("[")
            index = int(index)
            # obj = obj.setdefault(key, [{} for _ in range(index+1)])
            obj = obj.setdefault(key, [])
            while len(obj) <= index:
                obj.append({})
            obj = obj[index]
        else:
            obj = obj.setdefault(part, {})
    last_part = parts[-1]
    if "[" in last_part:
        key, index = last_part[:-1].split("[")
        index = int(index)
        if key not in obj or not isinstance(obj[key], list):
            obj[key] = [{} for _ in range(index + 1)]
        obj[key][index] = value
    else:
        obj[last_part] = value
    return ret_obj


def parse_update_expression(expr: str) -> Union[tuple[str, Any], None]:
    """
    Parse a string expression of the form 'path.to.field=value' into a path and a value.

    :param expr:
    :return:
    """
    try:
        path, val = expr.split("=", 1)
        val = json.loads(val)
    except ValueError:
        return None
    return path, val


def clean_empties(value: Union[Dict, List]) -> Any:
    if isinstance(value, dict):
        value = {k: v for k, v in ((k, clean_empties(v)) for k, v in value.items()) if v is not None}
    elif isinstance(value, list):
        value = [v for v in (clean_empties(v) for v in value) if v is not None]
    return value



from typing import List

from linkml_runtime import SchemaView
from linkml_runtime.linkml_model import SlotDefinition


def path_to_attribute_list(class_name: str, path: str, schema_view: SchemaView) -> List[SlotDefinition]:
    """
    Convert a path to a list of attributes.

    :param path:
    :return:
    """
    parts = path.split(".")
    att_list = []
    while parts:
        part = parts.pop(0)
        att = schema_view.induced_slot(part, class_name)
        if not att:
            raise ValueError(f"Attribute {part} not found in class {class_name}")
        att_list.append(att)
        class_name = att.range
    return att_list



import logging
from copy import copy
from typing import Any, Dict, List, Optional, Tuple, Union

from pymongo.collection import Collection as MongoCollection

from linkml_store.api import Collection
from linkml_store.api.collection import DEFAULT_FACET_LIMIT, OBJECT
from linkml_store.api.queries import Query, QueryResult

logger = logging.getLogger(__name__)


class MongoDBCollection(Collection):
    """
    Adapter for collections in a MongoDB database.

    .. note::

        You should not use or manipulate this class directly.
        Instead, use the general :class:`linkml_store.api.Collection`
    """

    @property
    def mongo_collection(self) -> MongoCollection:
        # collection_name = self.alias or self.name
        collection_name = self.alias
        if not collection_name:
            raise ValueError("Collection name not set")
        return self.parent.native_db[collection_name]

    def _check_if_initialized(self) -> bool:
        return self.alias in self.parent.native_db.list_collection_names()

    def insert(self, objs: Union[OBJECT, List[OBJECT]], **kwargs):
        if not isinstance(objs, list):
            objs = [objs]
        self.mongo_collection.insert_many(objs)
        # TODO: allow mapping of _id to id for efficiency
        for obj in objs:
            del obj["_id"]
        self._post_insert_hook(objs)

    def query(self, query: Query, limit: Optional[int] = None, offset: Optional[int] = None, **kwargs) -> QueryResult:
        mongo_filter = self._build_mongo_filter(query.where_clause)
        limit = limit or query.limit
        cursor = self.mongo_collection.find(mongo_filter)
        if limit and limit >= 0:
            cursor = cursor.limit(limit)
        offset = offset or query.offset
        if offset and offset >= 0:
            cursor = cursor.skip(offset)

        def _as_row(row: dict):
            row = copy(row)
            del row["_id"]
            return row

        rows = [_as_row(row) for row in cursor]
        count = self.mongo_collection.count_documents(mongo_filter)

        return QueryResult(query=query, num_rows=count, rows=rows)

    def _build_mongo_filter(self, where_clause: Dict[str, Any]) -> Dict[str, Any]:
        mongo_filter = {}
        if where_clause:
            for field, value in where_clause.items():
                mongo_filter[field] = value
        return mongo_filter

    from typing import Any, Dict, List, Union

    def query_facets(
        self,
        where: Dict = None,
        facet_columns: List[Union[str, Tuple[str, ...]]] = None,
        facet_limit=DEFAULT_FACET_LIMIT,
        **kwargs,
    ) -> Dict[Union[str, Tuple[str, ...]], List[Tuple[Any, int]]]:
        results = {}
        if not facet_columns:
            facet_columns = list(self.class_definition().attributes.keys())

        for col in facet_columns:
            logger.debug(f"Faceting on {col}")

            # Handle tuple columns
            if isinstance(col, tuple):
                group_id = {k.replace(".", "_"): f"${k}" for k in col}
                all_fields = col
            else:
                group_id = f"${col}"
                all_fields = [col]

            # Initial pipeline without unwinding
            facet_pipeline = [
                {"$match": where} if where else {"$match": {}},
                {"$group": {"_id": group_id, "count": {"$sum": 1}}},
                {"$sort": {"count": -1}},
                {"$limit": facet_limit},
            ]

            logger.info(f"Initial facet pipeline: {facet_pipeline}")
            initial_results = list(self.mongo_collection.aggregate(facet_pipeline))

            # Check if we need to unwind based on the results
            needs_unwinding = False
            if isinstance(col, tuple):
                needs_unwinding = any(
                    isinstance(result["_id"], dict) and any(isinstance(v, list) for v in result["_id"].values())
                    for result in initial_results
                )
            else:
                needs_unwinding = any(isinstance(result["_id"], list) for result in initial_results)

            if needs_unwinding:
                logger.info(f"Detected array values for {col}, unwinding...")
                facet_pipeline = [{"$match": where} if where else {"$match": {}}]

                # Unwind each field if needed
                for field in all_fields:
                    field_parts = field.split(".")
                    for i in range(len(field_parts)):
                        facet_pipeline.append({"$unwind": f"${'.'.join(field_parts[:i + 1])}"})

                facet_pipeline.extend(
                    [
                        {"$group": {"_id": group_id, "count": {"$sum": 1}}},
                        {"$sort": {"count": -1}},
                        {"$limit": facet_limit},
                    ]
                )

                logger.info(f"Updated facet pipeline with unwinding: {facet_pipeline}")
                facet_results = list(self.mongo_collection.aggregate(facet_pipeline))
            else:
                facet_results = initial_results

            logger.info(f"Facet results: {facet_results}")

            # Process results
            if isinstance(col, tuple):
                results[col] = [
                    (tuple(result["_id"].values()), result["count"])
                    for result in facet_results
                    if result["_id"] is not None and all(v is not None for v in result["_id"].values())
                ]
            else:
                results[col] = [
                    (result["_id"], result["count"]) for result in facet_results if result["_id"] is not None
                ]

        return results

    def delete(self, objs: Union[OBJECT, List[OBJECT]], **kwargs) -> int:
        if not isinstance(objs, list):
            objs = [objs]
        filter_conditions = []
        for obj in objs:
            filter_condition = {}
            for key, value in obj.items():
                filter_condition[key] = value
            filter_conditions.append(filter_condition)
        result = self.mongo_collection.delete_many({"$or": filter_conditions})
        return result.deleted_count

    def delete_where(self, where: Optional[Dict[str, Any]] = None, missing_ok=True, **kwargs) -> int:
        logger.info(f"Deleting from {self.target_class_name} where: {where}")
        if where is None:
            where = {}
        result = self.mongo_collection.delete_many(where)
        deleted_rows_count = result.deleted_count
        if deleted_rows_count == 0 and not missing_ok:
            raise ValueError(f"No rows found for {where}")
        return deleted_rows_count



from pathlib import Path
from typing import Iterable, Iterator, Optional, TextIO, Union

from linkml_runtime import SchemaView

from linkml_store.api.collection import OBJECT


def export_objects(
    objects: Iterable[OBJECT],
    location: Union[Path, str, TextIO],
    output_type: Optional[str],
    target_class: Optional[str] = None,
    schema_view: Optional[SchemaView] = None,
    **kwargs,
):
    """
    Export objects to a file or stream

    :param objects: objects to export
    :param location: location to export to
    :param kwargs:
    :return:
    """
    raise NotImplementedError


def import_objects_iter(
    location: Union[Path, str, TextIO], schema_view: Optional[SchemaView] = None, **kwargs
) -> Iterator[OBJECT]:
    """
    Import objects from a file or stream

    :param location:
    :param kwargs:
    :return:
    """
    raise NotImplementedError



import logging
from pathlib import Path
from typing import Dict, Optional, Union

import yaml
from linkml_runtime import SchemaView

from linkml_store.api import Database
from linkml_store.api.config import ClientConfig
from linkml_store.api.stores.chromadb.chromadb_database import ChromaDBDatabase
from linkml_store.api.stores.duckdb.duckdb_database import DuckDBDatabase
from linkml_store.api.stores.filesystem.filesystem_database import FileSystemDatabase
from linkml_store.api.stores.mongodb.mongodb_database import MongoDBDatabase
from linkml_store.api.stores.neo4j.neo4j_database import Neo4jDatabase
from linkml_store.api.stores.solr.solr_database import SolrDatabase

logger = logging.getLogger(__name__)


HANDLE_MAP = {
    "duckdb": DuckDBDatabase,
    "solr": SolrDatabase,
    "mongodb": MongoDBDatabase,
    "chromadb": ChromaDBDatabase,
    "neo4j": Neo4jDatabase,
    "file": FileSystemDatabase,
}


class Client:
    """
    A client is the top-level object for interacting with databases.

    * A client has access to one or more :class:`.Database` objects.
    * Each database consists of a number of :class:`.Collection` objects.

    Creating a client
    -----------------
    >>> client = Client()

    Attaching a database
    --------------------
    >>> db = client.attach_database("duckdb", alias="test")

    Note that normally a handle would be specified by a locator such as ``duckdb:///<PATH>``, but
    for convenience, an in-memory duckdb object can be specified without a full locator

    We can check the actual handle:

    >>> db.handle
    'duckdb:///:memory:'

    Creating a new collection
    -------------------------
    >>> collection = db.create_collection("Person")
    >>> objs = [{"id": "P1", "name": "John", "age_in_years": 30}, {"id": "P2", "name": "Alice", "age_in_years": 25}]
    >>> collection.insert(objs)
    >>> qr = collection.find()
    >>> len(qr.rows)
    2
    >>> qr.rows[0]["id"]
    'P1'
    >>> qr.rows[1]["name"]
    'Alice'
    >>> qr = collection.find({"name": "John"})
    >>> len(qr.rows)
    1
    >>> qr.rows[0]["name"]
    'John'

    """

    metadata: Optional[ClientConfig] = None
    _databases: Optional[Dict[str, Database]] = None

    def __init__(self, handle: Optional[str] = None, metadata: Optional[ClientConfig] = None):
        """
        Initialize a client.

        :param handle:
        :param metadata:
        """
        self.metadata = metadata
        if not self.metadata:
            self.metadata = ClientConfig()
        self.metadata.handle = handle

    @property
    def handle(self) -> Optional[str]:
        return self.metadata.handle

    @property
    def base_dir(self) -> Optional[str]:
        """
        Get the base directory for the client.

        Wraps metadata.base_dir.

        :return:
        """
        return self.metadata.base_dir

    def from_config(self, config: Union[ClientConfig, dict, str, Path], base_dir=None, **kwargs):
        """
        Create a client from a configuration.

        Examples
        --------
        >>> from linkml_store.api.config import ClientConfig
        >>> client = Client().from_config(ClientConfig(databases={"test": {"handle": "duckdb:///:memory:"}}))
        >>> len(client.databases)
        1
        >>> "test" in client.databases
        True
        >>> client.databases["test"].handle
        'duckdb:///:memory:'

        :param config:
        :param kwargs:
        :return:

        """
        if isinstance(config, dict):
            config = ClientConfig(**config)
        if isinstance(config, Path):
            config = str(config)
        if isinstance(config, str):
            # if not base_dir:
            #    base_dir = Path(config).parent
            parsed_obj = yaml.safe_load(open(config))
            config = ClientConfig(**parsed_obj)
        self.metadata = config
        if base_dir:
            self.metadata.base_dir = base_dir
        self._initialize_databases(**kwargs)
        return self

    def _initialize_databases(self, **kwargs):
        for name, db_config in self.metadata.databases.items():
            base_dir = self.base_dir
            logger.info(f"Initializing database: {name}, base_dir: {base_dir}")
            if not base_dir:
                base_dir = Path.cwd()
                logger.info(f"Using current working directory: {base_dir}")
            handle = db_config.handle.format(base_dir=base_dir)
            db_config.handle = handle
            if db_config.schema_location:
                db_config.schema_location = db_config.schema_location.format(base_dir=base_dir)
            db = self.attach_database(handle, alias=name, **kwargs)
            db.from_config(db_config)

    def attach_database(
        self,
        handle: str,
        alias: Optional[str] = None,
        schema_view: Optional[SchemaView] = None,
        recreate_if_exists=False,
        **kwargs,
    ) -> Database:
        """
        Associate a database with a handle.

        Examples
        --------
        >>> client = Client()
        >>> db = client.attach_database("duckdb", alias="memory")
        >>> "memory" in client.databases
        True
        >>> db = client.attach_database("duckdb:///tmp/another.db", alias="disk")
        >>> len(client.databases)
        2
        >>> "disk" in client.databases
        True

        :param handle: handle for the database, e.g. duckdb:///foo.db
        :param alias: alias for the database, e.g foo
        :param schema_view: schema view to associate with the database
        :param kwargs:
        :return:
        """
        if ":" not in handle:
            scheme = handle
            handle = None
            if alias is None:
                alias = scheme
        else:
            scheme, _ = handle.split(":", 1)
        if scheme not in HANDLE_MAP:
            raise ValueError(f"Unknown scheme: {scheme}")
        cls = HANDLE_MAP[scheme]
        db = cls(handle=handle, recreate_if_exists=recreate_if_exists, **kwargs)
        if schema_view:
            db.set_schema_view(schema_view)
        if not alias:
            alias = handle
        if not self._databases:
            self._databases = {}
        self._databases[alias] = db
        db.parent = self
        if db.alias:
            if db.alias != alias:
                raise AssertionError(f"Inconsistent alias: {db.alias} != {alias}")
        else:
            db.metadata.alias = alias
        return db

    def get_database(self, name: Optional[str] = None, create_if_not_exists=True, **kwargs) -> Database:
        """
        Get a named database.

        Examples
        --------
        >>> client = Client()
        >>> db = client.attach_database("duckdb:///test.db", alias="test")
        >>> retrieved_db = client.get_database("test")
        >>> db == retrieved_db
        True

        :param name: if None, there must be a single database attached
        :param create_if_not_exists:
        :param kwargs:
        :return:

        """
        if not name:
            if not self._databases:
                raise ValueError("No databases attached and no name provided")
            if len(self._databases) > 1:
                raise ValueError("Ambiguous: No name provided and multiple databases attached")
            return list(self._databases.values())[0]
        if not self._databases:
            self._databases = {}
        if name not in self._databases:
            if create_if_not_exists:
                logger.info(f"Creating database: {name}")
                self.attach_database(name, **kwargs)
            else:
                raise ValueError(f"Database {name} does not exist")
        return self._databases[name]

    @property
    def databases(self) -> Dict[str, Database]:
        """
        Return all attached databases

        Examples

        >>> client = Client()
        >>> _ = client.attach_database("duckdb", alias="test1")
        >>> _ = client.attach_database("duckdb", alias="test2")
        >>> len(client.databases)
        2
        >>> "test1" in client.databases
        True
        >>> "test2" in client.databases
        True
        >>> client.databases["test1"].handle
        'duckdb:///:memory:'
        >>> client.databases["test2"].handle
        'duckdb:///:memory:'

        :return:

        """
        if not self._databases:
            self._databases = {}
        return self._databases

    def drop_database(self, name: str, missing_ok=False, **kwargs):
        """
        Drop a database.

        Example (in-memory):

        >>> client = Client()
        >>> db1 = client.attach_database("duckdb", alias="test1")
        >>> db2 = client.attach_database("duckdb", alias="test2")
        >>> len(client.databases)
        2
        >>> client.drop_database("test1")
        >>> len(client.databases)
        1

        Databases that persist on disk:

        >>> client = Client()
        >>> path = Path("tmp/test.db")
        >>> path.parent.mkdir(parents=True, exist_ok=True)
        >>> db = client.attach_database(f"duckdb:///{path}", alias="test")
        >>> len(client.databases)
        1
        >>> db.store({"persons": [{"id": "P1", "name": "John"}]})
        >>> db.commit()
        >>> Path("tmp/test.db").exists()
        True
        >>> client.drop_database("test")
        >>> len(client.databases)
        0
        >>> Path("tmp/test.db").exists()
        False

        Dropping a non-existent database:

        >>> client = Client()
        >>> client.drop_database("duckdb:///tmp/made-up1", missing_ok=True)
        >>> client.drop_database("duckdb:///tmp/made-up2", missing_ok=False)
        Traceback (most recent call last):
        ...
        ValueError: Database duckdb:///tmp/made-up2 not found

        :param name:
        :param missing_ok:
        :return:
        """
        if self._databases:
            if name in self._databases:
                db = self._databases[name]
                db.drop(**kwargs)
                del self._databases[name]
            else:
                if not missing_ok:
                    raise ValueError(f"Database {name} not found")
        else:
            db = self.get_database(name, create_if_not_exists=True)
            db.drop(**kwargs)

    def drop_all_databases(self, **kwargs):
        """
        Drop all databases.

        Example (in-memory):

        >>> client = Client()
        >>> db1 = client.attach_database("duckdb", alias="test1")
        >>> assert "test1" in client.databases
        >>> db2 = client.attach_database("duckdb", alias="test2")
        >>> assert "test2" in client.databases
        >>> client.drop_all_databases()
        >>> len(client.databases)
        0


        :param missing_ok:
        :return:
        """
        if not self._databases:
            return
        for name in list(self._databases.keys()):
            self.drop_database(name, missing_ok=False, **kwargs)
        self._databases = {}



import logging
import sys
import warnings
from typing import Optional

import click
import yaml
from linkml_runtime.dumpers import json_dumper
from pydantic import BaseModel

from linkml_store import Client
from linkml_store.api import Collection, Database
from linkml_store.api.queries import Query
from linkml_store.index import get_indexer
from linkml_store.index.implementations.simple_indexer import SimpleIndexer
from linkml_store.index.indexer import Indexer
from linkml_store.utils.format_utils import Format, guess_format, load_objects, render_output, write_output
from linkml_store.utils.object_utils import object_path_update
from linkml_store.utils.pandas_utils import facet_summary_to_dataframe_unmelted

index_type_option = click.option(
    "--index-type",
    "-t",
    default="simple",
    show_default=True,
    help="Type of index to create. Values: simple, llm",
)

logger = logging.getLogger(__name__)

warnings.filterwarnings("ignore", module="duckdb_engine")


class ContextSettings(BaseModel):
    """
    Context object for CLI commands.
    """

    client: Client
    database_name: Optional[str] = None
    collection_name: Optional[str] = None

    @property
    def database(self) -> Optional[Database]:
        """
        Get the database object.
        :return:
        """
        name = self.database_name
        if name is None:
            # if len(self.client.databases) > 1:
            #    raise ValueError("Database must be specified if there are multiple databases.")
            if not self.client.databases:
                return None
            name = list(self.client.databases.keys())[0]
        return self.client.get_database(name)

    @property
    def collection(self) -> Optional[Collection]:
        """
        Get the collection object.
        :return:
        """
        name = self.collection_name
        if name is None:
            # if len(self.database.list_collections()) > 1:
            #    raise ValueError("Collection must be specified if there are multiple collections.")
            if not self.database.list_collections():
                return None
            name = list(self.database.list_collections())[0]
        return self.database.get_collection(name)

    class Config:
        arbitrary_types_allowed = True


# format_choice = click.Choice(["json", "yaml", "tsv"])
format_choice = click.Choice([f.value for f in Format])


include_internal_option = click.option("--include-internal/--no-include-internal", default=False, show_default=True)


@click.group()
@click.option("--database", "-d", help="Database name")
@click.option("--collection", "-c", help="Collection name")
@click.option("--config", "-C", type=click.Path(exists=True), help="Path to the configuration file")
@click.option("--set", help="Metadata settings in the form PATHEXPR=value", multiple=True)
@click.option("-v", "--verbose", count=True)
@click.option("-q", "--quiet/--no-quiet")
@click.option("--base-dir", "-B", help="Base directory for the client configuration")
@click.option(
    "--stacktrace/--no-stacktrace",
    default=False,
    show_default=True,
    help="If set then show full stacktrace on error",
)
@click.pass_context
def cli(ctx, verbose: int, quiet: bool, stacktrace: bool, database, collection, config, set, **kwargs):
    """A CLI for interacting with the linkml-store."""
    if not stacktrace:
        sys.tracebacklimit = 0
    logger = logging.getLogger()
    # Set handler for the root logger to output to the console
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s"))

    # Clear existing handlers to avoid duplicate messages if function runs multiple times
    logger.handlers = []

    # Add the newly created console handler to the logger
    logger.addHandler(console_handler)
    if verbose >= 2:
        logger.setLevel(logging.DEBUG)
    elif verbose == 1:
        logger.setLevel(logging.INFO)
    else:
        logger.setLevel(logging.WARNING)
    if quiet:
        logger.setLevel(logging.ERROR)
    ctx.ensure_object(dict)
    client = Client().from_config(config, **kwargs) if config else Client()
    settings = ContextSettings(client=client, database_name=database, collection_name=collection)
    ctx.obj["settings"] = settings
    # DEPRECATED
    ctx.obj["client"] = client
    ctx.obj["database"] = database
    ctx.obj["collection"] = collection
    if settings.database_name:
        db = client.get_database(database)
        if set:
            for expr in set:
                if "=" not in expr:
                    raise ValueError(f"Expression must be of form PARAM=VALUE. Got: {expr}")
                path, val = expr.split("=", 1)
                val = yaml.safe_load(val)
                logger.info(f"Setting {path} to {val}")
                db.metadata = object_path_update(db.metadata, path, val)
        # settings.database = db
        # DEPRECATED
        ctx.obj["database_obj"] = db
        if collection:
            collection_obj = db.get_collection(collection)
            ctx.obj["collection_obj"] = collection_obj
    if not settings.database_name:
        # if len(client.databases) != 1:
        #    raise ValueError("Database must be specified if there are multiple databases.")
        if client.databases:
            settings.database_name = list(client.databases.keys())[0]
    if not settings.collection_name:
        # if len(settings.database.list_collections()) != 1:
        #    raise ValueError("Collection must be specified if there are multiple collections.")
        if settings.database and settings.database.list_collections():
            collection = settings.database.list_collections()[0]
            settings.collection_name = collection.alias


@cli.command()
@click.argument("files", type=click.Path(exists=True), nargs=-1)
@click.option("--format", "-f", type=format_choice, help="Input format")
@click.option("--object", "-i", multiple=True, help="Input object as YAML")
@click.pass_context
def insert(ctx, files, object, format):
    """Insert objects from files (JSON, YAML, TSV) into the specified collection.

    Using a configuration:

        linkml-store -C config.yaml -c genes insert data/genes/*.json

    Note: if you don't provide a schema this will be inferred, but it is
    usually better to provide an explicit schema
    """
    settings = ctx.obj["settings"]
    collection = settings.collection
    if not collection:
        raise ValueError("Collection must be specified.")
    objects = []
    if not files and not object:
        files = ["-"]
    for file_path in files:
        if format:
            objects = load_objects(file_path, format=format)
        else:
            objects = load_objects(file_path)
        logger.info(f"Inserting {len(objects)} objects from {file_path} into collection '{collection.alias}'.")
        collection.insert(objects)
        click.echo(f"Inserted {len(objects)} objects from {file_path} into collection '{collection.alias}'.")
    if object:
        for object_str in object:
            logger.info(f"Parsing: {object_str}")
            objects = yaml.safe_load(object_str)
            collection.insert(objects)
            click.echo(f"Inserted {len(objects)} objects from {object_str} into collection '{collection.alias}'.")
    collection.commit()


@cli.command()
@click.argument("files", type=click.Path(exists=True), nargs=-1)
@click.option("--format", "-f", type=format_choice, help="Input format")
@click.option("--object", "-i", multiple=True, help="Input object as YAML")
@click.pass_context
def store(ctx, files, object, format):
    """Store objects from files (JSON, YAML, TSV) into the specified collection."""
    settings = ctx.obj["settings"]
    db = settings.database
    if not files and not object:
        files = ["-"]
    for file_path in files:
        if format:
            objects = load_objects(file_path, format=format)
        else:
            objects = load_objects(file_path)
        logger.info(f"Inserting {len(objects)} objects from {file_path} into database '{db}'.")
        for obj in objects:
            db.store(obj)
        click.echo(f"Inserted {len(objects)} objects from {file_path} into database '{db}'.")
    if object:
        for object_str in object:
            logger.info(f"Parsing: {object_str}")
            objects = yaml.safe_load(object_str)
            for obj in objects:
                db.store(obj)
            click.echo(f"Inserted {len(objects)} objects from {object_str} into collection '{db.name}'.")


@cli.command(name="import")
@click.option("--format", "-f", help="Input format")
@click.pass_context
@click.argument("files", type=click.Path(exists=True), nargs=-1)
def import_database(ctx, files, format):
    """Imports a database from a dump.

    See the `export` command for a full list of supported formats. The same
    formats are generally supported for imports.
    """
    settings = ctx.obj["settings"]
    db = settings.database
    if not files and not object:
        files = ["-"]
    for file_path in files:
        db.import_database(file_path, source_format=format)


@cli.command()
@click.option("--output-type", "-O", type=format_choice, default="json", help="Output format")
@click.option("--output", "-o", required=True, type=click.Path(), help="Output file path")
@click.pass_context
def export(ctx, output_type, output):
    """Exports a database to a standard dump format.

    Example:

        linkml-store -d duckdb:///countries.db export -O yaml -o countries.yaml

    Export format will be guessed from extension if not specified

    Example:

        linkml-store -d duckdb:///countries.db export -o countries.json

    Tree formats such as YAML and JSON can natively store an entire database; each collection
    will be a distinct key in the database.

    Additionally, native dump formats can be used:

    Example:

        linkml-store -d duckdb:///countries.db export -o countries -O duckdb

    Here, `countries` is a directory. This is equivalent to running EXPORT DATABASE
    (see https://duckdb.org/docs/sql/statements/export.html)
    """
    settings = ctx.obj["settings"]
    db = settings.database
    if output_type is None:
        output_type = guess_format(output)
    if output_type is None:
        raise ValueError(f"Output format must be specified can't be inferred from {output}.")
    db.export_database(output, target_format=output_type)


@cli.command()
@click.option("--output", "-o", type=click.Path(), help="Output file path")
@click.option("--output-type", "-O", type=format_choice, default="json", help="Output format")
@click.option("--other-database", "-D", required=False, help="Path to the other database")
@click.option("--other-collection", "-X", required=True, help="Name of the other collection")
@click.option("--identifier-attribute", "-I", required=False, help="Primary key name")
@click.pass_context
def diff(ctx, output, output_type, other_database, other_collection, identifier_attribute):
    """Diffs two collectoons to create a patch."""
    settings = ctx.obj["settings"]
    db = settings.database
    collection = settings.collection
    if not collection:
        raise ValueError("Collection must be specified.")
    other_db = settings.client.get_database(other_database) if other_database else db
    other_collection = other_db.get_collection(other_collection)
    if identifier_attribute:
        collection.set_identifier_attribute_name(identifier_attribute)
        other_collection.set_identifier_attribute_name(identifier_attribute)
    diff = collection.diff(other_collection)
    write_output(diff, output_type, target=output)


@cli.command()
@click.option("--identifier-attribute", "-I", required=False, help="Primary key name")
@click.argument("patch_files", type=click.Path(exists=True), nargs=-1)
@click.pass_context
def apply(ctx, patch_files, identifier_attribute):
    """
    Apply a patch to a collection.
    """
    settings = ctx.obj["settings"]
    collection = settings.collection
    if not collection:
        raise ValueError("Collection must be specified.")
    if identifier_attribute:
        collection.set_identifier_attribute_name(identifier_attribute)
    for patch_file in patch_files:
        patch_objs = load_objects(patch_file, expected_type=list)
        collection.apply_patches(patch_objs)


@cli.command()
@click.option("--where", "-w", type=click.STRING, help="WHERE clause for the query, as YAML")
@click.option("--limit", "-l", type=click.INT, help="Maximum number of results to return")
@click.option("--output-type", "-O", type=format_choice, default="json", help="Output format")
@click.option("--output", "-o", type=click.Path(), help="Output file path")
@click.pass_context
def query(ctx, where, limit, output_type, output):
    """Query objects from the specified collection.


    Leave the query field blank to return all objects in the collection.

    Examples:

        linkml-store -d duckdb:///countries.db -c countries query

    Queries can be specified in YAML, as basic key-value pairs

    Examples:

        linkml-store -d duckdb:///countries.db -c countries query -w 'code: NZ'

    More complex queries can be specified using MongoDB-style query syntax

    Examples:

        linkml-store -d file:. -c persons query  -w 'occupation: {$ne: Architect}'

    Finds all people who are not architects.
    """
    collection = ctx.obj["settings"].collection
    where_clause = yaml.safe_load(where) if where else None
    query = Query(from_table=collection.alias, where_clause=where_clause, limit=limit)
    result = collection.query(query)
    output_data = render_output(result.rows, output_type)
    if output:
        with open(output, "w") as f:
            f.write(output_data)
        click.echo(f"Query results saved to {output}")
    else:
        click.echo(output_data)


@cli.command()
@click.pass_context
@include_internal_option
def list_collections(ctx, **kwargs):
    db = ctx.obj["settings"].database
    for collection in db.list_collections(**kwargs):
        click.echo(collection.alias)
        click.echo(render_output(collection.metadata))


@cli.command()
@click.option("--where", "-w", type=click.STRING, help="WHERE clause for the query")
@click.option("--limit", "-l", type=click.INT, help="Maximum number of results to return")
@click.option("--output-type", "-O", type=format_choice, default="json", help="Output format")
@click.option("--output", "-o", type=click.Path(), help="Output file path")
@click.option("--columns", "-S", help="Columns to facet on")
@click.option("--wide/--no-wide", "-U/--no-U", default=False, show_default=True, help="Wide table")
@click.pass_context
def fq(ctx, where, limit, columns, output_type, wide, output):
    """
    Query facets from the specified collection.

    :param ctx:
    :param where:
    :param limit:
    :param columns:
    :param output_type:
    :param output:
    :return:
    """
    collection = ctx.obj["settings"].collection
    where_clause = yaml.safe_load(where) if where else None
    columns = columns.split(",") if columns else None
    if columns:
        columns = [col.strip() for col in columns]
        columns = [(tuple(col.split("+")) if "+" in col else col) for col in columns]
    logger.info(f"Faceting on columns: {columns}")
    results = collection.query_facets(where_clause, facet_columns=columns, limit=limit)
    logger.info(f"Facet results: {results}")

    def _untuple(key):
        if isinstance(key, tuple):
            return "+".join([str(x) for x in key])
        return key

    if wide:
        results_obj = facet_summary_to_dataframe_unmelted(results)
    else:
        if output_type == Format.PYTHON.value:
            results_obj = results
        elif output_type in [Format.TSV.value, Format.CSV.value]:
            results_obj = []
            for fc, data in results.items():
                for v, c in data:
                    results_obj.append({"facet": fc, "value": v, "count": c})
        else:
            results_obj = {}
            for key, value in results.items():
                value_as_dict = {_untuple(v[0:-1]): v[-1] for v in value}
                results_obj[_untuple(key)] = value_as_dict
    output_data = render_output(results_obj, output_type)
    if output:
        with open(output, "w") as f:
            f.write(output_data)
        click.echo(f"Query results saved to {output}")
    else:
        click.echo(output_data)


def _get_index(index_type=None, **kwargs) -> Indexer:
    if index_type is None or index_type == "simple":
        return SimpleIndexer(name="test", **kwargs)
    else:
        raise ValueError(f"Unknown index type: {index_type}")


@cli.command()
@click.option("--where", "-w", type=click.STRING, help="WHERE clause for the query")
@click.option("--output-type", "-O", type=format_choice, default=Format.FORMATTED.value, help="Output format")
@click.option("--output", "-o", type=click.Path(), help="Output file path")
@click.option(
    "--limit", "-l", default=-1, show_default=True, type=click.INT, help="Maximum number of results to return"
)
@click.pass_context
def describe(ctx, where, output_type, output, limit):
    """
    Describe the collection schema.
    """
    where_clause = yaml.safe_load(where) if where else None
    collection = ctx.obj["settings"].collection
    df = collection.find(where_clause, limit=limit).rows_dataframe
    write_output(df.describe(include="all").transpose(), output_type, target=output)


@cli.command()
@index_type_option
@click.option("--cached-embeddings-database", "-E", help="Path to the database where embeddings are cached")
@click.option("--text-template", "-T", help="Template for text embeddings")
@click.pass_context
def index(ctx, index_type, **kwargs):
    """
    Create an index over a collection.

    By default a simple trigram index is used.
    """
    collection = ctx.obj["settings"].collection
    ix = get_indexer(index_type, **kwargs)
    collection.attach_indexer(ix)


@cli.command()
@click.pass_context
@click.option("--output-type", "-O", type=format_choice, default="yaml", help="Output format")
@click.option("--output", "-o", type=click.Path(), help="Output file path")
def schema(ctx, output_type, output):
    """
    Show the schema for a database

    :param ctx:
    :param index_type:
    :return:
    """
    db = ctx.obj["settings"].database
    schema_dict = json_dumper.to_dict(db.schema_view.schema)
    output_data = render_output(schema_dict, output_type)
    if output:
        with open(output, "w") as f:
            f.write(output_data)
        click.echo(f"Schema saved to {output}")
    else:
        click.echo(output_data)


@cli.command()
@click.argument("search_term")
@click.option("--where", "-w", type=click.STRING, help="WHERE clause for the search")
@click.option("--limit", "-l", type=click.INT, help="Maximum number of search results")
@click.option("--output-type", "-O", type=format_choice, default="json", help="Output format")
@click.option("--output", "-o", type=click.Path(), help="Output file path")
@click.option(
    "--auto-index/--no-auto-index", default=False, show_default=True, help="Automatically index the collection"
)
@index_type_option
@click.pass_context
def search(ctx, search_term, where, limit, index_type, output_type, output, auto_index):
    """Search objects in the specified collection."""
    collection = ctx.obj["settings"].collection
    ix = get_indexer(index_type)
    logger.info(f"Attaching index to collection {collection.alias}: {ix.model_dump()}")
    collection.attach_indexer(ix, auto_index=auto_index)
    result = collection.search(search_term, where=where, limit=limit)
    output_data = render_output([{"score": row[0], **row[1]} for row in result.ranked_rows], output_type)
    if output:
        with open(output, "w") as f:
            f.write(output_data)
        click.echo(f"Search results saved to {output}")
    else:
        click.echo(output_data)


@cli.command()
@click.pass_context
def indexes(ctx):
    """
    Show the indexes for a collection.
    """
    collection = ctx.obj["settings"].collection
    for name, ix in collection.indexers.items():
        click.echo(f"{name}: {type(ix)}\n{ix.model_json()}")


@cli.command()
@click.option("--output-type", "-O", type=format_choice, default="json", help="Output format")
@click.option("--output", "-o", type=click.Path(), help="Output file path")
@click.pass_context
def validate(ctx, output_type, output):
    """Validate objects in the specified collection."""
    collection = ctx.obj["settings"].collection
    logger.info(f"Validating collection {collection.alias}")
    validation_results = [json_dumper.to_dict(x) for x in collection.iter_validate_collection()]
    output_data = render_output(validation_results, output_type)
    if output:
        with open(output, "w") as f:
            f.write(output_data)
        click.echo(f"Validation results saved to {output}")
    else:
        click.echo(output_data)


if __name__ == "__main__":
    cli()



import logging
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple, Union

from neo4j import Driver, Session

from linkml_store.api import Collection
from linkml_store.api.collection import DEFAULT_FACET_LIMIT, OBJECT
from linkml_store.api.queries import Query, QueryResult
from linkml_store.graphs.graph_map import EdgeProjection, GraphProjection, NodeProjection

logger = logging.getLogger(__name__)


class DeletePolicy(Enum):
    CASCADE = "cascade"
    ERROR = "error"
    STUB = "stub"


class Neo4jCollection(Collection):
    """
    Adapter for collections in a Neo4j database.
    """

    # _graph_projection: Optional[GraphProjection] = None
    delete_policy: DeletePolicy = DeletePolicy.CASCADE

    @property
    def driver(self) -> Driver:
        return self.parent.driver

    def session(self) -> Session:
        return self.parent.session()

    def _check_if_initialized(self) -> bool:
        with self.session() as session:
            result = session.run("MATCH (n) RETURN count(n) > 0 as exists")
            return result.single()["exists"]

    @property
    def graph_projection(self) -> GraphProjection:
        return self.metadata.graph_projection

    @property
    def node_projection(self) -> Optional[NodeProjection]:
        return self.metadata.graph_projection if isinstance(self.graph_projection, NodeProjection) else None

    @property
    def edge_projection(self) -> Optional[EdgeProjection]:
        return self.metadata.graph_projection if isinstance(self.graph_projection, EdgeProjection) else None

    @property
    def is_edge_collection(self) -> bool:
        return isinstance(self.graph_projection, EdgeProjection)

    @property
    def category_labels_attribute(self) -> str:
        np = self.node_projection
        category_labels_attribute = None
        if np:
            category_labels_attribute = np.category_labels_attribute
        if not category_labels_attribute:
            category_labels_attribute = "category"
        return category_labels_attribute

    @property
    def identifier_attribute(self) -> str:
        gp = self.graph_projection
        id_attribute = None
        if gp:
            id_attribute = gp.identifier_attribute
        if not id_attribute:
            id_attribute = "id"
        return id_attribute

    def _node_pattern(self, obj: Optional[OBJECT] = None, node_var="n") -> str:
        obj = {} if obj is None else obj
        category_labels_attribute = self.category_labels_attribute
        categories = obj.get(category_labels_attribute or "category", [])
        if not isinstance(categories, list):
            categories = [categories]
        cstr = (":" + ":".join(categories)) if categories else ""
        return f"{node_var}{cstr}"

    @property
    def is_node_collection(self) -> bool:
        return not self.is_edge_collection

    def set_is_edge_collection(self, force=False):
        if self.is_edge_collection:
            return
        if self.graph_projection and not force:
            raise ValueError("Cannot reassign without force=True")
        self.metadata.graph_projection = EdgeProjection()

    def set_is_node_collection(self, force=False):
        if self.is_node_collection:
            return
        if self.graph_projection and not force:
            raise ValueError("Cannot reassign without force=True")
        self.metadata.graph_projection = NodeProjection()

    def _prop_clause(
        self, obj: OBJECT, node_var: Optional[str] = None, exclude_attributes: Optional[List[str]] = None
    ) -> str:
        if exclude_attributes is None:
            exclude_attributes = [self.category_labels_attribute]
        node_prefix = node_var + "." if node_var else ""
        terms = [f"{node_prefix}{k}: ${k}" for k in obj.keys() if k not in exclude_attributes]
        return ", ".join(terms)

    def insert(self, objs: Union[OBJECT, List[OBJECT]], **kwargs):
        if not isinstance(objs, list):
            objs = [objs]
        self._pre_insert_hook(objs)

        with self.session() as session:
            for obj in objs:
                query = self._create_insert_cypher_query(obj)
                session.run(query, **obj)

        self._post_insert_hook(objs)

    def _create_insert_cypher_query(self, obj: OBJECT) -> str:
        id_attribute = self.identifier_attribute
        if not self.is_edge_collection:
            logger.debug(f"Inserting node: {obj}")
            category_labels_attribute = self.category_labels_attribute
            node_pattern = self._node_pattern(obj)
            props = self._prop_clause(obj, exclude_attributes=[id_attribute, category_labels_attribute])
            return f"CREATE ({node_pattern} {{{id_attribute}: ${id_attribute}, {props}}})"
        else:
            logger.debug(f"Inserting edge: {obj}")
            ep = self.edge_projection
            if ep.predicate_attribute not in obj:
                raise ValueError(f"Predicate attribute {ep.predicate_attribute} not found in edge {obj}.")
            if ep.subject_attribute not in obj:
                raise ValueError(f"Subject attribute {ep.subject_attribute} not found in edge {obj}.")
            if ep.object_attribute not in obj:
                raise ValueError(f"Object attribute {ep.object_attribute} not found in edge {obj}.")
            pred = obj[ep.predicate_attribute]
            # check if nodes present; if not, make dangling stubs
            # TODO: decide on how this should be handled in validation if some fields are required
            for node_id in [obj[ep.subject_attribute], obj[ep.object_attribute]]:
                check_query = (
                    f"MATCH (n {{{ep.identifier_attribute}: ${ep.identifier_attribute}}}) RETURN count(n) as count"
                )
                with self.session() as session:
                    result = session.run(check_query, **{ep.identifier_attribute: node_id})
                    if result.single()["count"] == 0:
                        if self.delete_policy == DeletePolicy.STUB:
                            stub_query = f"CREATE (n {{{ep.identifier_attribute}: ${ep.identifier_attribute}}})"
                            session.run(stub_query, **{ep.identifier_attribute: node_id})
                        else:
                            raise ValueError(f"Node with identifier {node_id} not found in the database.")
            edge_props = self._prop_clause(
                obj, exclude_attributes=[ep.subject_attribute, ep.predicate_attribute, ep.object_attribute]
            )
            return f"""
            MATCH (s {{{id_attribute}: ${ep.subject_attribute}}}), (o {{{id_attribute}: ${ep.object_attribute}}})
            CREATE (s)-[r:{pred} {{{edge_props}}}]->(o)
            """

    def _prop_clause(self, obj: OBJECT, exclude_attributes: List[str] = None, node_var: Optional[str] = None) -> str:
        if exclude_attributes is None:
            exclude_attributes = []
        node_prefix = f"{node_var}." if node_var else ""
        terms = [f"{node_prefix}{k}: ${k}" for k in obj.keys() if k not in exclude_attributes]
        return ", ".join(terms)

    def query(self, query: Query, limit: Optional[int] = None, offset: Optional[int] = None, **kwargs) -> QueryResult:
        cypher_query = self._build_cypher_query(query, limit, offset)
        ca = self.category_labels_attribute
        with self.session() as session:
            result = session.run(cypher_query, query.where_clause)
            if self.is_edge_collection:
                rows = [self._edge_to_dict(record) for record in result]
            else:

                def node_to_dict(n) -> dict:
                    d = dict(n.items())
                    if ca:
                        labels = list(n.labels)
                        if labels:
                            d[ca] = labels[0]
                    return d

                rows = [node_to_dict(record["n"]) for record in result]

        # count_query = self._build_count_query(query, is_count=True)
        count_query = self._build_cypher_query(query, is_count=True)
        with self.session() as session:
            count = session.run(count_query, query.where_clause).single()["count"]

        return QueryResult(query=query, num_rows=count, rows=rows)

    def _build_cypher_query(
        self, query: Query, limit: Optional[int] = None, offset: Optional[int] = None, is_count=False
    ) -> str:
        if self.is_edge_collection:
            ep = self.edge_projection
            ia = ep.identifier_attribute
            sa = ep.subject_attribute
            pa = ep.predicate_attribute
            oa = ep.object_attribute
            wc = query.where_clause or {}
            rq = "r"
            pred = wc.get(pa, None)
            if pred:
                rq = f"r:{pred}"
            sq = "s"
            subj = wc.get(sa, None)
            if subj:
                sq = f"s {{{ia}: '{subj}'}}"
            oq = "o"
            obj = wc.get(oa, None)
            if obj:
                oq = f"o {{{ia}: '{obj}'}}"
            where = {k: v for k, v in wc.items() if k not in [sa, pa, oa]}
            cypher_query = f"""
            MATCH ({sq})-[{rq}]->({oq})
            {self._build_where_clause(where, 'r')}
            """
            if is_count:
                cypher_query += """
                RETURN count(r) as count
                """
            else:
                cypher_query += f"""
                RETURN r, type(r) as predicate, s.{ia} as subject, o.{ia} as object
                """
        else:
            node_pattern = self._node_pattern(query.where_clause)
            cypher_query = f"""
            MATCH ({node_pattern})
            {self._build_where_clause(query.where_clause)}
            """
            if is_count:
                cypher_query += """
                RETURN count(n) as count
                """
            else:
                cypher_query += """
                RETURN n
                """

        if not is_count:
            if limit and limit >= 0:
                cypher_query += f" LIMIT {limit}"
            if offset and offset >= 0:
                cypher_query += f" SKIP {offset}"

        return cypher_query

    def _build_where_clause(self, where_clause: Dict[str, Any], prefix: str = "n") -> str:
        conditions = []
        if where_clause is None:
            return ""
        for key, value in where_clause.items():
            if key == self.category_labels_attribute:
                continue
            if isinstance(value, str):
                conditions.append(f"{prefix}.{key} = '{value}'")
            else:
                conditions.append(f"{prefix}.{key} = {value}")

        return "WHERE " + " AND ".join(conditions) if conditions else ""

    def _edge_to_dict(self, record: Dict) -> Dict[str, Any]:
        r = record["r"]
        ep = self.edge_projection
        return {
            ep.subject_attribute: record["subject"],
            ep.predicate_attribute: record["predicate"],
            ep.object_attribute: record["object"],
            **dict(r.items()),
        }

    def query_facets(
        self,
        where: Dict = None,
        facet_columns: List[Union[str, Tuple[str, ...]]] = None,
        facet_limit=DEFAULT_FACET_LIMIT,
        **kwargs,
    ) -> Dict[Union[str, Tuple[str, ...]], List[Tuple[Any, int]]]:
        results = {}
        if not facet_columns:
            facet_columns = list(self.class_definition().attributes.keys())

        category_labels_attribute = self.category_labels_attribute
        with self.session() as session:
            for col in facet_columns:
                where_clause = self._build_where_clause(where) if where else ""
                if col == category_labels_attribute:
                    # Handle faceting on labels
                    query = f"""
                    MATCH (n)
                    {where_clause}
                    WITH labels(n) AS nodeLabels, count(*) as count
                    UNWIND nodeLabels AS label
                    WITH label, count
                    ORDER BY count DESC, label
                    LIMIT {facet_limit}
                    RETURN label as value, count
                    """
                else:
                    query = f"""
                    MATCH (n)
                    {where_clause}
                    WITH n.{col} as value, count(*) as count
                    WITH value, count
                    ORDER BY count DESC
                    LIMIT {facet_limit}
                    RETURN value, count
                    """
                result = session.run(query)
                results[col] = [(record["value"], record["count"]) for record in result]

        return results

    def delete(self, objs: Union[OBJECT, List[OBJECT]], **kwargs) -> int:
        delete_policy = self.delete_policy
        if not isinstance(objs, list):
            objs = [objs]

        deleted_nodes = 0
        deleted_relationships = 0
        identifier_attribute = self.identifier_attribute

        with self.session() as session:
            for obj in objs:
                node_pattern = self._node_pattern(obj)
                id_value = obj[identifier_attribute]
                where_clause = f"{{{identifier_attribute}: $id}}"
                dn, dr = self._execute_delete(session, node_pattern, where_clause, delete_policy, id=id_value)
                deleted_nodes += dn
                deleted_relationships += dr

        return deleted_nodes

    def delete_where(self, where: Optional[Dict[str, Any]] = None, missing_ok=True, **kwargs) -> int:
        delete_policy = self.delete_policy
        where_clause = self._build_where_clause(where) if where else ""
        node_pattern = self._node_pattern(where)

        with self.session() as session:
            deleted_nodes, deleted_relationships = self._execute_delete(
                session, node_pattern, where_clause, delete_policy
            )

        if deleted_nodes == 0 and not missing_ok:
            raise ValueError(f"No nodes found for {where}")

        return deleted_nodes

    def _execute_delete(
        self, session, node_pattern: str, where_clause: str, delete_policy: DeletePolicy, **params
    ) -> Tuple[int, int]:
        deleted_relationships = 0
        deleted_nodes = 0

        if delete_policy == DeletePolicy.ERROR:
            check_query = f"MATCH ({node_pattern} {where_clause})-[r]-() RETURN count(r) as rel_count"
            result = session.run(check_query, **params)
            if result.single()["rel_count"] > 0:
                raise ValueError("Nodes with existing relationships found and cannot be deleted.")

        if delete_policy == DeletePolicy.CASCADE:
            rel_query = f"MATCH ({node_pattern} {where_clause})-[r]-() DELETE r"
            result = session.run(rel_query, **params)
            deleted_relationships = result.consume().counters.relationships_deleted

        if delete_policy in [DeletePolicy.CASCADE, DeletePolicy.ERROR]:
            node_query = f"MATCH ({node_pattern} {where_clause}) DELETE n"
            result = session.run(node_query, **params)
            deleted_nodes = result.consume().counters.nodes_deleted
        elif delete_policy == DeletePolicy.STUB:
            stub_query = f"MATCH ({node_pattern} {where_clause}) SET n.deleted = true RETURN count(n) as stub_count"
            result = session.run(stub_query, **params)
            deleted_nodes = result.single()["stub_count"]

        return deleted_nodes, deleted_relationships

    def update(self, objs: Union[OBJECT, List[OBJECT]], **kwargs) -> int:
        if not isinstance(objs, list):
            objs = [objs]

        updated_count = 0
        with self.session() as session:
            for obj in objs:
                query = self._create_update_cypher_query(obj)
                result = session.run(query, **obj)
                updated_count += result.consume().counters.properties_set

        return updated_count

    def _create_update_cypher_query(self, obj: OBJECT) -> str:
        id_attribute = self.identifier_attribute
        category_labels_attribute = self.category_labels_attribute

        # Prepare SET clause
        set_items = [f"n.{k} = ${k}" for k in obj.keys() if k not in [id_attribute, category_labels_attribute]]
        set_clause = ", ".join(set_items)

        # Prepare labels update
        labels_to_add = []
        # labels_to_remove = []
        if category_labels_attribute in obj:
            new_labels = (
                obj[category_labels_attribute]
                if isinstance(obj[category_labels_attribute], list)
                else [obj[category_labels_attribute]]
            )
            labels_to_add = [f":{label}" for label in new_labels]
            # labels_to_remove = [":Label" for _ in new_labels]  # Placeholder for labels to remove

        # Construct the query
        query = f"MATCH (n {{{id_attribute}: ${id_attribute}}})\n"
        # f labels_to_remove:
        #    query += f"REMOVE n{' '.join(labels_to_remove)}\n"
        if labels_to_add:
            query += f"SET n{' '.join(labels_to_add)}\n"
            # f"REMOVE n{' '.join(labels_to_remove)}' if labels_to_remove else ''}"
            # f"{f'SET n{' '.join(labels_to_add)}' if labels_to_add else ''}"
        query += f"SET {set_clause}\n"
        query += "RETURN n"
        print(query)
        return query
